#!/usr/bin/env python
"""Tests for `fast_dash` package."""
# pylint: disable=redefined-outer-name

from fast_dash import FastDash, dcc, dbc, dmc, html, Chat
from fast_dash.Components import Text, Slider, UploadImage
from fast_dash.utils import _pil_to_b64, Fastify

from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC

import time
import datetime


########### Define callback functions ###########
def simple_text_to_text_function(input_text: Text):
    "Converts text to text"
    return input_text


def simple_number_to_number(input_: Slider):
    return input_

def simple_image_to_image(input_: UploadImage):
    return UploadImage


def simple_text_to_multiple_text_function(input_text):
    return input_text, input_text


############## Write test cases ################
def test_fdco001_input_is_fc(dash_duo):
    "When the input is a FastComponent"

    # Input type is Text
    app = FastDash(callback_fn=simple_text_to_text_function)
    assert app.inputs_with_ids[0].__doc__ == Text.__doc__, "Input type is Text failed"

    # Input type is Slider
    app = FastDash(callback_fn=simple_number_to_number, inputs=Slider)
    assert app.inputs_with_ids[0].__doc__ == Slider.__doc__, "Input type is Slider failed"

    # Input type is Image
    app = FastDash(callback_fn=simple_image_to_image, inputs=UploadImage)
    assert app.inputs_with_ids[0].__doc__ == UploadImage.__doc__, "Input type is Image failed"


def test_fdco002_hint_is_not_of_type_type(dash_duo):
    "Input is not of type `type`"

    # Hint is text
    def simple_text(text: "Fast Dash"):
        return text

    app = FastDash(callback_fn=simple_text)
    input_component = app.inputs_with_ids[0]
    assert input_component.__doc__ == dmc.Textarea().__doc__ and not hasattr(
        input_component, "type"
    ), "Hint is text failed"

    # Hint is integer
    def simple_integer(integer: 5):
        return integer

    app = FastDash(callback_fn=simple_integer)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dbc.Input().__doc__
        and input_component.type == "number"
    ), "Hint is integer failed"


def test_fdco003_input_is_image(dash_duo):
    "When hint and default values are independently PIL images"

    # Input hint is PIL image
    import PIL
    from PIL import ImageFile

    def simple_image(image: PIL.ImageFile.ImageFile):
        return image

    app = FastDash(callback_fn=simple_image)
    input_component = app.inputs_with_ids[0]
    assert input_component.__doc__ == UploadImage.__doc__, "Hint is PIL image failed"

    # Defualt value is image
    from PIL import Image
    import requests
    from io import BytesIO

    url = "https://storage.googleapis.com/fast_dash/0.2.7/Mosaic%20examples/ex1.png"
    response = requests.get(url)
    img = Image.open(BytesIO(response.content))

    def simple_image(image=img):
        return image

    app = FastDash(callback_fn=simple_image)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == UploadImage.__doc__
        and input_component.ack.__doc__ == html.Img.__doc__
    ), "Defualt value is image failed"


def test_fdco004_input_hint_is_text(dash_duo):
    "Try different default types when the input hint is text"

    # 1. Default is text
    def simple_text(text: str = "Default text"):
        return text

    app = FastDash(callback_fn=simple_text)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == Text.__doc__
        and hasattr(input_component, "value")
        and input_component.value == "Default text"
    ), "Default text failed"

    # 2. Default is numeric
    def simple_text(text: str = 2.2):
        return text

    app = FastDash(callback_fn=simple_text)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dbc.Input().__doc__
        and hasattr(input_component, "type")
        and input_component.type == "number"
        and hasattr(input_component, "value")
        and input_component.value == 2.2
    ), "Default number failed"

    # 3. Default is sequence
    def simple_text(text: str = ["Some text", 2.2, "45.23"]):
        return text

    app = FastDash(callback_fn=simple_text)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dcc.Dropdown().__doc__
        and hasattr(input_component, "options")
        and input_component.options == ["Some text", 2.2, "45.23"]
    ), "Default sequence failed"

    # 4. Default is dictionary
    def simple_text(
        text: str = {"This": "is", "a": "dictionary", 5: "Fast", "Dash": [1, 2, 3]}
    ):
        return text

    app = FastDash(callback_fn=simple_text)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dbc.Input().__doc__
        and not hasattr(input_component, "type")
        and hasattr(input_component, "value")
        and input_component.value
        == str({"This": "is", "a": "dictionary", 5: "Fast", "Dash": [1, 2, 3]})
    ), "Default dictionary failed"

    # 5. Default is boolean
    def simple_text(text: str = True):
        return text

    app = FastDash(callback_fn=simple_text)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dbc.Input().__doc__
        and not hasattr(input_component, "type")
        and hasattr(input_component, "value")
        and input_component.value == str(True)
    ), "Default boolean failed"

    # 5. Default is date
    def simple_text(text: str = datetime.date(2022, 8, 11)):
        return text

    app = FastDash(callback_fn=simple_text)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dbc.Input().__doc__
        and not hasattr(input_component, "type")
        and hasattr(input_component, "value")
        and input_component.value == str(datetime.date(2022, 8, 11))
    ), "Default date failed"

    # 6. Default is datetime
    def simple_text(text: str = datetime.datetime(2022, 8, 11, 12, 44, 56)):
        return text

    app = FastDash(callback_fn=simple_text)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dbc.Input().__doc__
        and not hasattr(input_component, "type")
        and hasattr(input_component, "value")
        and input_component.value == str(datetime.datetime(2022, 8, 11, 12, 44, 56))
    ), "Default datetime failed"


def test_fdco005_input_hint_is_numeric(dash_duo):
    "Try different default types when the input hint is numeric"

    # 1. Default is text
    def simple_num(num: float = "2.2"):
        return num

    app = FastDash(callback_fn=simple_num)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dbc.Input().__doc__
        and hasattr(input_component, "type")
        and input_component.type == "number"
        and hasattr(input_component, "value")
        and input_component.value == 2.2
    ), "Default text failed"

    # 2. Default is numeric
    def simple_num(num: float = 2.2):
        return num

    app = FastDash(callback_fn=simple_num)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dbc.Input().__doc__
        and hasattr(input_component, "type")
        and input_component.type == "number"
        and hasattr(input_component, "value")
        and input_component.value == 2.2
    ), "Default number failed"

    # 3a. Default value is of type range
    def simple_num(num: float = range(1, 10, 2)):
        return num

    app = FastDash(callback_fn=simple_num)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dcc.Slider().__doc__
        and hasattr(input_component, "min")
        and input_component.min == 1
        and hasattr(input_component, "max")
        and input_component.max == 10
        and hasattr(input_component, "step")
        and input_component.step == 2
    ), "Default range failed"

    # 3b. Default value is other sequence (list)
    def simple_num(num: float = [0, 2, 3, 4, 5, 6, 7, 80]):
        return num

    app = FastDash(callback_fn=simple_num)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dcc.Slider().__doc__
        and hasattr(input_component, "min")
        and input_component.min == 0
        and hasattr(input_component, "max")
        and input_component.max == 80
        and hasattr(input_component, "step")
        and input_component.step == 10
    ), "Default other sequence failed"


def test_fdco006_input_hint_is_sequence(dash_duo):
    "Try different default types when the input hint is a sequence"

    # 1. Default is text
    def simple_list(l_: list = "This, is,a,list, 2.2"):
        return l_

    app = FastDash(callback_fn=simple_list)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dmc.MultiSelect().__doc__
        and not hasattr(input_component, "type")
        and hasattr(input_component, "value")
        and input_component.value == ["This, is,a,list, 2.2"]
    ), "Default text failed"

    # 2. Default is a list (sequence)
    def simple_list(l_: list = ["This", "is", "a", "list", "2.2"]):
        return l_

    app = FastDash(callback_fn=simple_list)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dmc.MultiSelect().__doc__
        and not hasattr(input_component, "type")
        and hasattr(input_component, "data")
        and input_component.data == ["This", "is", "a", "list", "2.2"]
        and hasattr(input_component, "component_property")
        and input_component.component_property == "value"
    ), "Default list failed"

    # 3. Default is a dictionary (sequence)
    sample_dictionary = {"This": "is", "a": "dictionary", 5: "Fast", "Dash": [1, 2, 3]}

    def simple_list(l_: list = sample_dictionary):
        return l_

    app = FastDash(callback_fn=simple_list)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dmc.MultiSelect().__doc__
        and not hasattr(input_component, "type")
        and hasattr(input_component, "data")
        and input_component.data == sample_dictionary
        and hasattr(input_component, "component_property")
        and input_component.component_property == "value"
    ), "Default dictionary failed"

    # 4. No default value
    def simple_list(l_: list):
        return l_

    app = FastDash(callback_fn=simple_list)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dmc.MultiSelect().__doc__
        and not hasattr(input_component, "type")
        and hasattr(input_component, "component_property")
        and input_component.component_property == "value"
    ), "No default failed"


def test_fdco007_input_hint_is_dictionary(dash_duo):
    "When the input hint is dictionary"

    # 1. Default is dictionary
    sample_dictionary = {"This": "is", "a": "dictionary", 5: "Fast", "Dash": [1, 2, 3]}

    def simple_dictionary(l_: dict = sample_dictionary):
        return l_

    app = FastDash(callback_fn=simple_dictionary)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dcc.Dropdown().__doc__
        and not hasattr(input_component, "type")
        and hasattr(input_component, "options")
        and input_component.options == sample_dictionary
        and hasattr(input_component, "component_property")
        and input_component.component_property == "value"
    ), "Default dictionary failed"


def test_fdco008_input_hint_is_boolean(dash_duo):
    "When the input hint is bool"

    # 1. Default is True
    def simple_bool(l_: bool = True):
        return l_

    app = FastDash(callback_fn=simple_bool)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dbc.Checkbox().__doc__
        and hasattr(input_component, "value")
        and input_component.value is True
        and hasattr(input_component, "component_property")
        and input_component.component_property == "value"
    ), "Default = True failed"

    # No default value (False by default)
    def simple_bool(l_: bool):
        return l_

    app = FastDash(callback_fn=simple_bool)
    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == dbc.Checkbox().__doc__
        and not hasattr(input_component, "value")
        and hasattr(input_component, "component_property")
        and input_component.component_property == "value"
    ), "Default = False failed"


def test_fdco009_input_hint_is_image(dash_duo):
    "When the input hint is image"

    # 1. Default value is PIL image
    import PIL
    from PIL import Image
    import requests
    from io import BytesIO

    url = "https://storage.googleapis.com/fast_dash/0.2.7/Mosaic%20examples/ex1.png"
    response = requests.get(url)
    img = Image.open(BytesIO(response.content))

    def simple_img(img_: PIL.ImageFile.ImageFile = img):
        return img_

    app = FastDash(callback_fn=simple_img)
    dash_duo.start_server(app.app)
    dash_duo.wait_for_text_to_equal("#title8888928", "Simple Img", timeout=4)
    time.sleep(4)

    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == UploadImage.__doc__
        and hasattr(input_component, "component_property")
        and input_component.component_property == "contents"
        and hasattr(input_component, "ack")
        and input_component.ack.__doc__ == html.Img.__doc__
        and dash_duo.find_element("#img__ack").get_attribute("src") == _pil_to_b64(img)
    ), "Default image failed"


def test_fdco010_input_hint_is_image_no_default(dash_duo):
    "Hint is of type image and no defautl value"

    # 1. No default value
    import PIL

    def simple_img(img_: PIL.ImageFile.ImageFile):
        return img_

    app = FastDash(callback_fn=simple_img)
    dash_duo.start_server(app.app)
    dash_duo.wait_for_text_to_equal("#title8888928", "Simple Img", timeout=4)
    time.sleep(4)

    input_component = app.inputs_with_ids[0]
    assert (
        input_component.__doc__ == UploadImage.__doc__
        and hasattr(input_component, "component_property")
        and input_component.component_property == "contents"
        and hasattr(input_component, "ack")
        and input_component.ack.__doc__ == html.Img.__doc__
        and dash_duo.find_element("#img__ack").get_attribute("src") == None
    ), "No default value failed"


def test_fdco011_input_hint_is_date(dash_duo):
    "Hint is of type date"

    # 1 Default value is date
    def simple_date(date: datetime.date = datetime.date(2022, 5, 4)):
        return date

    app = FastDash(callback_fn=simple_date)
    input_component = app.inputs_with_ids[0]

    assert (
        input_component.__doc__ == dcc.DatePickerSingle.__doc__
        and hasattr(input_component, "component_property")
        and input_component.component_property == "date"
        and hasattr(input_component, "date")
        and input_component.date == datetime.date(2022, 5, 4)
    ), "Default date failed"

    # 2. No default value
    def simple_date(date: datetime.date):
        return date

    app = FastDash(callback_fn=simple_date)
    input_component = app.inputs_with_ids[0]

    assert (
        input_component.__doc__ == dcc.DatePickerSingle.__doc__
        and hasattr(input_component, "component_property")
        and input_component.component_property == "date"
        and hasattr(input_component, "date")
        and input_component.date == datetime.date.today()
    ), "No default value failed"


def test_fdco012_input_hint_is_unknown(dash_duo):
    "When the input hint is unknown or unsupported"

    # 1. Default value is text
    def simple_text(text="Some text"):
        return text

    app = FastDash(callback_fn=simple_text)
    input_component = app.inputs_with_ids[0]

    assert (
        input_component.__doc__ == dbc.Input().__doc__
        and not hasattr(input_component, "type")
        and hasattr(input_component, "value")
        and input_component.value == "Some text"
    ), "Default text failed"

    # 2. Default value is a number
    def simple_num(text=137.2736):
        return text

    app = FastDash(callback_fn=simple_num)
    input_component = app.inputs_with_ids[0]

    assert (
        input_component.__doc__ == dbc.Input().__doc__
        and hasattr(input_component, "type")
        and input_component.type == "number"
        and hasattr(input_component, "value")
        and input_component.value == 137.2736
    ), "Default number failed"

    # 3a. Default value is a sequence (list)
    def simple_list(l_=["These", "are", 5, "options", "to select"]):
        return l_

    app = FastDash(callback_fn=simple_list)
    input_component = app.inputs_with_ids[0]

    assert (
        input_component.__doc__ == dcc.Dropdown().__doc__
        and hasattr(input_component, "options")
        and input_component.options == ["These", "are", 5, "options", "to select"]
    ), "Default list failed"

    # 3b. Default value is range
    def simple_range(l_=range(1, 1000, 5)):
        return l_

    app = FastDash(callback_fn=simple_range)
    input_component = app.inputs_with_ids[0]

    assert (
        input_component.__doc__ == dcc.Slider().__doc__
        and hasattr(input_component, "min")
        and input_component.min == 1
        and hasattr(input_component, "max")
        and input_component.max == 1000
        and hasattr(input_component, "step")
        and input_component.step == 5
    ), "Default range failed"

    # 4. Default value is a dictionary
    def simple_dict(l_={"This": "is", "a": "dictionary", 5: "Fast", "Dash": [1, 2, 3]}):
        return l_

    app = FastDash(callback_fn=simple_dict)
    input_component = app.inputs_with_ids[0]

    assert (
        input_component.__doc__ == dcc.Dropdown().__doc__
        and hasattr(input_component, "options")
        and input_component.options
        == {"This": "is", "a": "dictionary", 5: "Fast", "Dash": [1, 2, 3]}
    ), "Default dictionary failed"

    # 5. Default value is boolean
    def simple_bool(l_=True):
        return l_

    app = FastDash(callback_fn=simple_bool)
    input_component = app.inputs_with_ids[0]

    assert (
        input_component.__doc__ == dbc.Checkbox().__doc__
        and hasattr(input_component, "component_property")
        and input_component.component_property == "value"
        and hasattr(input_component, "value")
        and input_component.value is True
    ), "Default boolean failed"

    # 6. Default value is date
    def simple_date(l_=datetime.date(2022, 12, 25)):
        return l_

    app = FastDash(callback_fn=simple_date)
    input_component = app.inputs_with_ids[0]

    assert (
        input_component.__doc__ == dcc.DatePickerSingle.__doc__
        and hasattr(input_component, "component_property")
        and input_component.component_property == "date"
        and hasattr(input_component, "date")
        and input_component.date == datetime.date(2022, 12, 25)
    ), "Default date failed"

    # 7. Default value is timestamp (currently treated as date)
    def simple_datetime(l_=datetime.datetime(2022, 12, 25, 14, 55, 35)):
        return l_

    app = FastDash(callback_fn=simple_datetime)
    input_component = app.inputs_with_ids[0]

    assert (
        input_component.__doc__ == dcc.DatePickerSingle.__doc__
        and hasattr(input_component, "component_property")
        and input_component.component_property == "date"
        and hasattr(input_component, "date")
        and input_component.date == datetime.date(2022, 12, 25)
    ), "Default timestamp failed"

    # 8. Unsupported default value type
    def simple_unsupported(inp_=time.localtime()):
        return inp_

    app = FastDash(callback_fn=simple_unsupported)
    input_component = app.inputs_with_ids[0]

    assert input_component.__doc__ == dmc.Textarea().__doc__ and not hasattr(
        input_component, "type"
    ), "Default unsupported failed"

    # 9. Default input type is `None`
    def simple_unsupported(inp_=None):
        return inp_

    app = FastDash(callback_fn=simple_unsupported)
    input_component = app.inputs_with_ids[0]

    assert input_component.__doc__ == dmc.Textarea().__doc__ and not hasattr(
        input_component, "type"
    ), "Default None failed"


def test_fdco013_output_hint_is_fast_component(dash_duo):
    "When the output hint type is a Fast Component"

    def simple_number(num) -> Slider:
        return num

    app = FastDash(callback_fn=simple_number)
    output_component = app.outputs[0]

    assert (
        output_component.__doc__ == dcc.Slider().__doc__
        and hasattr(output_component, "min")
        and output_component.min == 0
        and hasattr(output_component, "max")
        and output_component.max == 20
        and hasattr(output_component, "step")
        and output_component.step == 1
    ), "Default FC failed"


def test_fdco014_output_hint_is_image(dash_duo):
    "When the output hint type is an image or type(image)"

    # 1. Output is subclass of PIL.ImageFile.ImageFile
    import PIL

    def simple_image(img) -> PIL.ImageFile.ImageFile:
        return img

    app = FastDash(callback_fn=simple_image)
    output_component = app.outputs[0]

    assert (
        output_component.__doc__ == html.Img.__doc__
        and hasattr(output_component, "component_property")
        and output_component.component_property == "src"
    ), "Default PIL.ImageFile.ImageFile failed"

    # 2. Output is an image object
    from PIL import Image
    import requests
    from io import BytesIO

    url = "https://storage.googleapis.com/fast_dash/0.2.7/Mosaic%20examples/ex1.png"
    response = requests.get(url)
    img = Image.open(BytesIO(response.content))

    def simple_image(img_) -> img:
        return img_

    app = FastDash(callback_fn=simple_image)
    output_component = app.outputs[0]

    assert (
        output_component.__doc__ == html.Img.__doc__
        and hasattr(output_component, "component_property")
        and output_component.component_property == "src"
    ), "Default image failed"

    # 3. Output is of type `type(image)`
    img_type = type(img)

    def simple_image(img_) -> img_type:
        return img_

    app = FastDash(callback_fn=simple_image)
    output_component = app.outputs[0]

    assert (
        output_component.__doc__ == html.Img.__doc__
        and hasattr(output_component, "component_property")
        and output_component.component_property == "src"
    ), "Default type(image) failed"


def test_fdco015_input_is_dash_component(dash_duo):
    "When the input hint type is a Dash component"

    # Component property is specified
    component = Fastify(dmc.Text(), "children")

    def simple_text_to_text(text: component):
        return "Some markdown text"

    app = FastDash(callback_fn=simple_text_to_text)
    input_component = app.inputs[0]

    assert (
        input_component.__doc__ == dmc.Text.__doc__
        and hasattr(input_component, "component_property")
        and input_component.component_property == "children"
    ), "Fast component failed"

    # Component property is not specified
    def simple_text_to_text(text: dmc.Text()):
        return "Some markdown text"

    app = FastDash(callback_fn=simple_text_to_text)
    input_component = app.inputs[0]

    assert (
        input_component.__doc__ == dmc.Text.__doc__
        and hasattr(input_component, "component_property")
        and input_component.component_property == "children"
    ), "Dash component failed"


def test_fdco016_output_is_dash_component(dash_duo):
    "When the output hint type is a Dash component"

    # Component property is specified
    component = Fastify(dcc.Markdown(), "children")

    def simple_text_to_text(text) -> component:
        return "Some markdown text"

    app = FastDash(callback_fn=simple_text_to_text)
    output_component = app.outputs[0]

    assert (
        output_component.__doc__ == dcc.Markdown.__doc__
        and hasattr(output_component, "component_property")
        and output_component.component_property == "children"
    ), "Fast component failed"

    # Component property is not specified
    def simple_text_to_text(text) -> dcc.Markdown():
        return "Some markdown text"

    app = FastDash(callback_fn=simple_text_to_text)
    output_component = app.outputs[0]

    assert (
        output_component.__doc__ == dcc.Markdown.__doc__
        and hasattr(output_component, "component_property")
        and output_component.component_property == "children"
    ), "Dash component failed"


def test_fdco017_output_is_chat(dash_duo):
    "When the output hint type is a Chat Fast component"

    def simple_chat(query: str) -> Chat:
        response = f"Response to {query}"
        chat = {"query": query, "response": response}
        
        return chat
    
    app = FastDash(callback_fn=simple_chat).app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal("#title8888928", "Simple Chat", timeout=4)

    # Enter some text
    form_textfield = dash_duo.find_element("#query")
    form_textfield.send_keys("Why?")

    # Click submit
    dash_duo.multiple_click("#submit_inputs", 1)

    # Check if any child element has the text "Response to Why?"
    output_div = dash_duo.find_element("#output-1")

    wait = WebDriverWait(dash_duo.driver, timeout=4)
    wait.until(EC.presence_of_element_located((By.CSS_SELECTOR, "#output-1")))
    time.sleep(4)

    child_elements = output_div.find_elements(By.CSS_SELECTOR, "*")
    text_found = False

    for element in child_elements:
        if "Response to Why?" in element.text:
            text_found = True
            break

    assert text_found, "Response text not found in any child element of output-1"


def test_fdco018_output_is_pandas(dash_duo):
    "When the output hint type is a Chat Fast component"

    import pandas as pd
    def simple_table(dummy_input: int) -> pd.DataFrame:
        df = pd.DataFrame(data={"Col1": range(100), 
                                "Col2": [f"C{i}" for i in range(100)],
                                "Date": pd.date_range(start="2023-01-01", periods=100)})
        
        return df

    app = FastDash(callback_fn=simple_table).app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal("#title8888928", "Simple Table", timeout=4)

    # Click submit
    dash_duo.multiple_click("#submit_inputs", 1)
    time.sleep(4)

    # Ensure the table is present
    table = dash_duo.find_element("#output-1")
    assert table is not None
    
    # Validate the data (example: check the first cell)
    first_cell = table.find_element_by_css_selector(".dash-cell div")
    assert first_cell.text == "0"  # Adjust according to expected content


from selenium.webdriver.chrome.options import Options


def pytest_setup_options():
    options = Options()
    options.add_argument("--disable-gpu")
    options.add_argument("--headless")
    options.add_argument("--window-size=1280x1024")
    options.add_argument("--disable-pytest-warnings")
    return options



from fast_dash import FastDash, Fastify
from fast_dash.Components import Text, Image, Upload, UploadImage, Slider, html, dcc
from fast_dash.utils import _pil_to_b64


def example_1_simple_text_to_text():
    "Fast Dash example 1. Simple text to text Fast Dash app"
    ## Define callback functions
    def simple_text_to_text_function(input_text):
        return input_text

    app = FastDash(
        callback_fn=simple_text_to_text_function,
        inputs=Text,
        outputs=Text,
        title="Fast Dash example 1",
    )

    return app


def example_2_text_with_slider():
    "Fast Dash example 2"

    # Step 1: Define your model inference
    def text_to_text_function(input_text, slider_value):
        processed_text = f"{input_text}. Slider value is {slider_value}."
        return processed_text

    # Step 2: Specify the input and output components
    app = FastDash(
        callback_fn=text_to_text_function,
        inputs=[Text, Slider],
        outputs=Text,
        title="Fast Dash example 2",
    )

    return app


def example_3_image_to_image():
    "Fast Dash example 3"

    # Step 1: Define your model inference
    def callback_fn(image):
        return image

    # Step 2: Specify the input and output components
    app = FastDash(
        callback_fn=callback_fn,
        inputs=Upload,
        outputs=Image,
        title="Fast Dash example 3",
    )

    return app


def example_4_image_slider_to_image_text():
    "Fast Dash example 4. Input is Upload (with ack) and slider. Output is Image and Text."

    def callback_fn(input_text, slider_value):
        return input_text, f"Slider value is {slider_value}"

    ack_image = Fastify(html.Img(width="100%"), "src")
    fast_upload = Fastify(
        dcc.Upload(
            children=["Click to upload"],
            style={"borderStyle": "dashed", "padding-bottom": "20px"},
        ),
        "contents",
        ack=ack_image,
    )

    app = FastDash(
        callback_fn=callback_fn,
        inputs=[fast_upload, Slider],
        outputs=[Image, Text],
        title="Fast Dash example 4",
        theme="SKETCHY",
    )

    return app


def example_5_uploadimage_to_image():
    "Fast Dash example 5. Input is UploadImage. Output is Image."

    def image_to_image(image):

        from PIL import Image
        import io
        import base64

        _, image_contents = image.split(",")
        processed_image = Image.open(
            io.BytesIO(base64.b64decode(image_contents.encode()))
        )

        return _pil_to_b64(processed_image)

    app = FastDash(
        callback_fn=image_to_image,
        inputs=UploadImage,
        outputs=Image,
        title="Fast Dash example 5",
        title_image_path="https://storage.googleapis.com/fast_dash/0.2.7/Mosaic%20examples/ex1.png",
        subheader="Build ML prototypes lightning fast!",
        github_url="https://github.com/dkedar7/fast_dash/",
        linkedin_url="https://linkedin.com/in/dkedar7/",
        twitter_url="https://twitter.com/dkedar7/",
        theme="FLATLY",
    )

    return app


def example_6_text_to_plt():
    "Fast Dash example 5. Input is UploadImage. Output is Image."

    import matplotlib.pyplot as plt

    def text_to_plt(some_text) -> plt.Figure:

        fig, ax = plt.subplots(1, 1)
        ax.plot([1, 2, 3], [4, 5, 6])

        return fig

    app = FastDash(
        callback_fn=text_to_plt,
        inputs=Text,
        outputs=Image,
        title="Fast Dash example 6",
        disable_logs=True
    )

    return app



"""Top-level package for Fast Dash."""

__author__ = """Kedar Dabhadkar"""
__email__ = "kedar@fastdash.app"
__version__ = "0.2.9"

from fast_dash.Components import (
    Graph,
    Image,
    Slider,
    Text,
    TextArea,
    Upload,
    UploadImage,
    acknowledge_image_component,
    dbc,
    dcc,
    dmc,
    html,
    PIL,
    Chat,
    Table
)
import dash
from fast_dash.fast_dash import FastDash, fastdash
from fast_dash.utils import Fastify

__all__ = [
    "FastDash",
    "fastdash",
    "Fastify",
    "Text",
    "TextArea",
    "Slider",
    "Upload",
    "acknowledge_image_component",
    "UploadImage",
    "Image",
    "Graph",
    "dcc",
    "dbc",
    "dmc",
    "html",
    "dash",
    "PIL",
    "Chat",
    "Table"
]



import copy
import datetime
import inspect
import math
import numbers
import warnings
from collections.abc import Iterable, Sequence
from functools import reduce

import dash
from dash import Input, Output, State, dcc, html, ctx, Patch
from dash.exceptions import PreventUpdate
from flask import request

import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
from dash_iconify import DashIconify

import matplotlib as mpl
import numpy as np
import pandas as pd
import PIL
from PIL import ImageFile
import plotly.graph_objs as go

from .utils import (
    Fastify,
    _pil_to_b64,
    _get_default_property,
    _parse_docstring_as_markdown,
)


class BaseLayout:
    def __init__(
        self,
        mosaic=None,
        inputs=None,
        outputs=None,
        title=None,
        title_image_path=None,
        subtitle=None,
        github_url=None,
        linkedin_url=None,
        twitter_url=None,
        navbar=True,
        footer=True,
        about=True,
        minimal=False,
        scale_height=1,
        app=None,
    ):
        self.mosaic = mosaic
        self.inputs = inputs
        self.outputs = outputs
        self.title = title
        self.title_image_path = title_image_path
        self.subtitle = subtitle
        self.github_url = github_url
        self.linkedin_url = linkedin_url
        self.twitter_url = twitter_url
        self.navbar = navbar
        self.footer = footer
        self.about = about
        self.minimal = minimal
        self.scale_height = scale_height
        self.app = app

    def generate_navbar_container(self):
        if not self.navbar:
            return None

        navbar_components = []

        # 1. Add navbar components
        if self.about:
            navbar_components.append(
                dbc.Row(dbc.Button("About", color="primary", id="about-navlink"))
            )

            # Add modal component
            navbar_components.append(
                dmc.Modal(id="about-modal", size="80%", zIndex=10000)
            )

        if self.github_url:
            navbar_components.append(
                dmc.NavLink(
                    label=DashIconify(icon="ri:github-fill", color="#ffffff", width=30),
                    href=self.github_url,
                    target="_blank",
                    variant="light",
                )
            )

        if self.linkedin_url:
            navbar_components.append(
                dmc.NavLink(
                    label=DashIconify(
                        icon="entypo-social:linkedin-with-circle",
                        color="#ffffff",
                        width=30,
                    ),
                    href=self.linkedin_url,
                    target="_blank",
                    variant="light",
                )
            )

        if self.twitter_url:
            navbar_components.append(
                dmc.NavLink(
                    label=DashIconify(
                        icon="formkit:twitter", color="#ffffff", width=30
                    ),
                    href=self.twitter_url,
                    target="_blank",
                    variant="light",
                )
            )

        navbar = dbc.NavbarSimple(
            children=[] + navbar_components,
            brand=[
                dmc.Group(
                    [
                        dmc.Burger(
                            id="sidebar-button", opened=True, color="white", size=15
                        ),
                        self.title,
                    ],
                    spacing=5,
                )
            ]
            or "",
            color="primary",
            dark=True,
            fluid=True,
            fixed=None,
            expand=True,
            style={"padding": "0 0 0 0"},
        )

        navbar_container = dbc.Row(
            [navbar],
            # fluid=True,
            style={"padding": "0 0 0 0"},
            id="navbar3260780",
            # className="dbc",
        )

        return navbar_container

    def generate_header_component(self):
        header_children = []

        if self.title:
            header_children.append(
                dbc.Row(
                    html.H2(
                        self.title, style={"textAlign": "center"}, id="title8888928"
                    ),
                    style={"padding": "1% 0% 1% 0%"},
                )
            )

        if self.title_image_path:
            header_children.append(
                dbc.Row(
                    dbc.Row(
                        html.Img(src=self.title_image_path, style={"width": "250px"}),
                        justify="center",
                    )
                )
            )

        if self.subtitle:
            header_children.append(
                dbc.Row(
                    dbc.Row(
                        html.H5(
                            self.subtitle,
                            id="subheader6904007",
                            style={"textAlign": "center"},
                        ),
                        style={"padding": "0% 0% 1% 0%"},
                    )
                )
            )

        header_container = dbc.Container(header_children, id="header1162572")

        return header_container

    def generate_input_component(self):
        input_container = dbc.Col(
            children=self.inputs,
            id="input-group",
            style={
                "padding": "2% 1% 1% 2%",
                "lineHeight": "60px",
                "borderWidth": "1px",
                "borderRadius": "5px",
                "background-color": "#F2F2F2",
            },
            width=5,
            xs=12,
            sm=12,
            md=5,
            lg=5,
            xl=5,
            xxl=5,
        )

        return input_container

    def generate_output_component(self):
        output_container = dbc.Col(
            children=self.outputs,
            id="output-group",
            style={
                "padding": "2% 1% 1% 2%",
                "lineHeight": "60px",
                "borderWidth": "1px",
                "borderRadius": "5px",
                "background-color": "#F2F2F2",
            },
            width=5,
            xs=12,
            sm=12,
            md=5,
            lg=5,
            xl=5,
            xxl=5,
        )

        return output_container

    def generate_footer_container(self):
        if self.footer is False:
            return None

        footer = dbc.NavbarSimple(
            brand="Made with Fast Dash",
            brand_href="https://fastdash.app/",
            color="primary",
            dark=True,
            fluid=True,
            fixed=None,
            style={"padding": "0 0 0 0"},
        )

        footer = html.A(
            "Made with Fast Dash", href="https://fastdash.app/", target="_blank"
        )

        footer_container = dbc.Container(
            [footer], fluid=True, style={"padding": "0 0 0 2%"}, id="footer5265971"
        )

        return footer_container

    def generate_layout(self):
        if self.minimal:
            self.title = self.subtitle = self.navbar = self.footer = False

        layout = dbc.Container(
            [
                html.Div(id="dummy-div", style={"display": "none"}),
                self.generate_navbar_container(),
                self.generate_header_component(),
                dbc.Row(
                    [self.generate_input_component(), self.generate_output_component()],
                    justify="evenly",
                    style={"padding": "2% 1% 0% 2%"},
                ),
                self.generate_footer_container(),
            ],
            fluid=True,
            style={"padding": "0 0 0 0"},
        )

        return layout

    def callbacks(self, app):
        "Optional callbacks specific to the layout"

        @app.app.callback(
            Output("about-modal", "opened"),
            Output("about-modal", "children"),
            Input("about-navlink", "n_clicks"),
            State("about-modal", "opened"),
        )
        def display_function_about_information(n_clicks, opened):
            if n_clicks:
                return not opened, app.callback_fn.__doc__

            raise PreventUpdate


class SidebarLayout(BaseLayout):
    def __init__(self, **kwargs):
        self.unique_components = []

        self.col_style = {"text-align": "center"}
        self.row_style = {"text-align": "center"}

        super().__init__(**kwargs)

        if self.mosaic is None:
            self.mosaic = self._infer_mosaic(self.outputs)

    @staticmethod
    def _infer_mosaic(outputs):
        # Number of outputs
        n = len(outputs) - 1

        def calc_lcm(a, b):
            # Calculate LCM
            return abs(a * b) // math.gcd(a, b)

        def lcm_list(numbers):
            # LCM of numbers in a list
            return reduce(calc_lcm, numbers)

        def split_into_chunks(n):
            # Split into chunks of equal sizes
            full_chunks = n // 3
            remainder = n % 3
            result = [3] * full_chunks
            if remainder != 0:
                result.append(remainder)
            return result

        alphabets = list("ABCDEFGHIJKLMNOPQRSTUVWXYZ")

        # Split into chunks
        chunks = split_into_chunks(n)

        # Find LCM
        lcm = lcm_list(chunks)

        # Loop over alphanets to generate the final mosaic
        mosaic = ""
        for len_row in chunks:
            mosaic_row_list = []

            # Number of times to repeat elements in each row
            repetitions = int(lcm / len_row)

            for i in range(len_row):
                letter = alphabets.pop(0)
                mosaic_row_list += [letter] * repetitions

            mosaic += "".join(mosaic_row_list) + "\n"

        mosaic = mosaic.strip()

        return mosaic

    @staticmethod
    def _normalize_grid_string(layout):
        if "\n" not in layout:
            # single-line string
            return [list(ln) for ln in layout.split(";")]
        else:
            # multi-line string
            layout = inspect.cleandoc(layout)
            return [list(ln) for ln in layout.strip("\n").split("\n")]

    @staticmethod
    def _make_array(inp):
        """
        Convert input into 2D array

        Returns
        -------
        2D object array
        """
        r0, *rest = inp
        if isinstance(r0, str):
            raise ValueError("List mosaic specification must be 2D")
        for j, r in enumerate(rest, start=1):
            if isinstance(r, str):
                raise ValueError("List mosaic specification must be 2D")
            if len(r0) != len(r):
                raise ValueError(
                    "All of the rows must be the same length, however "
                    f"the first row ({r0!r}) has length {len(r0)} "
                    f"and row {j} ({r!r}) has length {len(r)}."
                )
        out = np.zeros((len(inp), len(r0)), dtype=object)
        for j, r in enumerate(inp):
            for k, v in enumerate(r):
                out[j, k] = v
        return out

    @staticmethod
    def _check_if_rectangular(mosaic):
        "Returns True if all elements in the mosaic array make a perfect rectangular area"

        # Get all the unique keys
        unique_ids = np.unique(mosaic)

        # go through the unique keys,
        for name in unique_ids:
            # sort out where each axes starts/ends
            indx = np.argwhere(mosaic == name)
            start_row, start_col = np.min(indx, axis=0)
            end_row, end_col = np.max(indx, axis=0) + 1

            # and construct the slice object
            slc = (slice(start_row, end_row), slice(start_col, end_col))

            # some light error checking
            if (mosaic[slc] != name).any():
                raise ValueError(
                    f"While trying to layout\n{mosaic!r}\n"
                    f"we found that the label {name!r} specifies a "
                    "non-rectangular or non-contiguous area."
                )

        return True

    @staticmethod
    def _get_n_unique(arr, axis):
        "Get unique number of elements in an array along a given axis"
        if axis not in [0, 1]:
            raise ValueError("Axis for a 2D array must be either 0 or 1")

        nunique_list = []

        for i in range(arr.shape[1 - axis]):
            sub_arr = arr[:, i] if axis == 0 else arr[i, :]
            nunique = np.unique(sub_arr).shape[0]
            nunique_list.append(nunique)

        return nunique_list

    @staticmethod
    def _get_break_indices(arr, axis, index):
        "Get the indices where an array elements switches to another"

        values = arr[:, index] if axis == 0 else arr[index, :]
        break_indices = []
        widths = []

        for idx, v in enumerate(values):
            if idx == 0:
                previous_break_index = 0
                previous_value = v
            else:
                if v != previous_value:
                    widths.append(idx - previous_break_index)
                    break_indices.append(idx)

                    previous_break_index = idx
                    previous_value = v

        widths.append(idx + 1 - previous_break_index)
        widths = [int((w * 12) / sum(widths)) for w in widths]
        extra_assign = [1 for i in range(12 % sum(widths))]
        extra_assign = extra_assign + [
            0 for i in range(len(widths) - len(extra_assign))
        ]

        widths = [
            int((w * 12) / sum(widths)) + ea for w, ea in zip(widths, extra_assign)
        ]

        return break_indices, widths

    @staticmethod
    def _get_sub_mosaics(arr, axis, index_breaks):
        "Get sub-mosaics along the given axis"

        sub_mosaics = []
        begin_index = 0
        index_breaks.append(9999)
        for i_break in index_breaks:
            if axis == 1:
                sub_mosaics.append(arr[:, begin_index:i_break])

            elif axis == 0:
                sub_mosaics.append(arr[begin_index:i_break, :])

            begin_index = i_break

        return sub_mosaics

    def _get_component(self, axis, width, style=None):
        if axis == 1:
            style = self.col_style if style is None else style
            style.update({"height": "100%"})
            layout = dbc.Col(
                [],
                style=style,
                width=width,
                class_name="p-0 d-flex flex-column flex-fill",
            )

        else:
            style = self.row_style if style is None else style
            # style.update({"max-height": "100%"})
            layout = dbc.Row([], style=style, class_name="p-0 g-1 d-flex flex-fill")

        return copy.deepcopy(layout)

    def _set_single_component(self, axis, width, n_rows=1, style=None, label=""):
        style = self.col_style if style is None else style
        component = self.output_component_mapper.get(label, label)
        style.update({"height": f"{n_rows * self.height_of_single_row}vh"})
        layout = dbc.Col(
            [component],
            class_name="p-1 bg-white flex-fill d-flex flex-column",
            style=style,
            width=width,
            # align="center",
        )

        return copy.deepcopy(layout)

    def _do_mosaic(self, mosaic_array, axis, layout):
        n_unique = len(np.unique(mosaic_array))

        if n_unique == 1:
            label = np.unique(mosaic_array)[0]

            self.unique_components.append(label)
            style = self.col_style

            lo = self._set_single_component(
                axis=axis,
                width=layout.width if isinstance(layout, dbc.Col) else None,
                style=style,
                label=label,
                n_rows=mosaic_array.shape[0],
            )
            return copy.deepcopy(lo)

        n_unique_axis = self._get_n_unique(arr=mosaic_array, axis=1 - axis)
        min_index = np.argmin(n_unique_axis)

        index_breaks, widths = self._get_break_indices(
            arr=mosaic_array, axis=1 - axis, index=min_index
        )
        sub_arrays = self._get_sub_mosaics(
            arr=mosaic_array, axis=1 - axis, index_breaks=index_breaks
        )

        for array, width in zip(sub_arrays, widths):
            child_layout = self._do_mosaic(
                array, 1 - axis, self._get_component(axis=1 - axis, width=width)
            )
            layout.children.append(child_layout)

        return layout

    def generate_input_component(self):
        return dbc.Col(
            children=[dmc.Stack(children=self.inputs)],
            id="input-group",
            xs=12,
            md=2,
            style={
                "background-color": "#F5F7F7",
                "display": "block",
                "padding": "2% 20px 0 20px",
                "height": f"{self.scale_height * 110}vh",
            },
            class_name="border border-right",
        )

    def generate_output_component(self):
        mosaic = self._normalize_grid_string(self.mosaic)
        mosaic_arr = self._make_array(mosaic)
        mosaic_shape = mosaic_arr.shape
        self.height_of_single_row = (80 * self.scale_height) / (mosaic_shape[0])

        # Check if the mosaic array makes rectangles for all elements
        self._check_if_rectangular(mosaic_arr)

        if self.outputs == []:
            self.output_component_mapper = {}

        else:
            unique_locations = np.unique(mosaic_arr)
            unique_locations.sort()
            self.output_component_mapper = {
                m: o for m, o in zip(unique_locations, self.outputs[:-1])
            }

        # Slice along the axis with the most number of unique elements first
        begin_axis = np.argmax(
            [
                max([len(np.unique(arr)) for arr in mosaic_arr]),
                max([len(np.unique(arr)) for arr in mosaic_arr.transpose()]),
            ]
        )

        begin = dbc.Row([], justify=True, class_name="g-1 d-flex")
        layout = self._do_mosaic(mosaic_arr, axis=1 - begin_axis, layout=begin)
        output_layout = dmc.LoadingOverlay(
            dbc.Col(
                [layout] + [self.outputs[-1]],
                class_name="g-1 d-flex flex-fill flex-column",
                style={"height": f"{80 * self.scale_height}vh"},
                width=12,
            ),
            loaderProps=dict(variant="bars"),
        )

        return output_layout

    def generate_footer_container(self):
        return dmc.Affix(
            dmc.Tooltip(
                label="Made with Fast Dash!",
                position="top",
                withArrow=True,
                transitionDuration=300,
                children=dcc.Link(
                    dmc.Button(
                        DashIconify(icon="ion:rocket-sharp", width=20), radius=500
                    ),
                    href="https://github.com/dkedar7/fast_dash",
                    target="_blank",
                ),
            ),
            position={"bottom": "20px", "right": "20px"},
            id="footer5265971",
        )

    def generate_layout(self):
        # There are four main components:
        # navbar, header, input, output, footer

        if self.minimal:
            self.title = self.subtitle = self.navbar = self.footer = False

        layout = dmc.MantineProvider(
            dmc.NotificationsProvider(
                [
                    html.Div(id="dummy-div", style={"display": "none"}),
                    dbc.Container(
                        [
                            self.generate_navbar_container(),
                            dbc.Row(
                                [
                                    self.generate_input_component(),
                                    dbc.Col(
                                        [
                                            self.generate_header_component(),
                                            self.generate_output_component(),
                                        ],
                                        id="output-group-col",
                                        style={"padding": "1% 2% 0 2%"},
                                    ),
                                ],
                                class_name="d-flex",
                            ),
                            self.generate_footer_container(),
                            html.Div(id="error-notify-div"),
                        ],
                        fluid=True,
                        style={"height": "100vh", "width": "100%"},
                    )
                ]
            )
        )

        return layout

    def callbacks(self, app):
        @app.app.callback(
            [Output("input-group", "style"), Output("sidebar-button", "opened")],
            [Input("sidebar-button", "opened")],
            [State("input-group", "style")],
        )
        def toggle_sidebar(opened, input_style):
            user_agent = request.headers.get("User-Agent")
            input_style = {} if input_style is None else input_style

            # Condition to collapse the sidebar:
            # Burger icon is closed or no inputs are specified
            if not opened or self.app.inputs == [] or self.app.inputs is None:
                input_style.update({"display": "none"})
                opened = False

            # Another condition to collapse the sidebar:
            # User is on mobile and the inputs are submitted
            elif ctx.triggered_id == "submit_inputs" and "Mobi" in user_agent:
                input_style.update({"display": "none"})
                opened = False

            # Expand by default
            else:
                input_style.update({"display": "block"})

            return (input_style, opened)

        # Optional callbacks specific to the layout
        @app.app.callback(
            Output("about-modal", "opened"),
            Output("about-modal", "children"),
            Input("about-navlink", "n_clicks"),
            State("about-modal", "opened"),
        )
        def display_function_about_information(n_clicks, opened):
            if n_clicks:
                if self.about == True:
                    about_text = _parse_docstring_as_markdown(
                        app.callback_fn, title=self.title
                    )

                elif isinstance(self.about, str):
                    about_text = self.about

                else:
                    about_text = _parse_docstring_as_markdown(
                        app.callback_fn, title=self.title
                    )

                return not opened, dcc.Markdown(about_text)

            raise PreventUpdate


def _get_readable_names_from_parent_classes(type_hint):
    "Get a readable label for the object's type. Order is important. If disturbed, type = bool could get matched with float."
    _map_types_to_readable_names = {
        str: "Text",
        bool: "Boolean",
        float: "Numeric",
        int: "Numeric",
        complex: "Numeric",
        list: "Sequence",
        set: "Sequence",
        range: "Sequence",
        # tuple: "Sequence",
        PIL.Image.Image: "Image",
        dict: "Dictionary",
        datetime.datetime: "Timestamp",
        datetime.date: "Date",
    }

    for parent_class in _map_types_to_readable_names:
        if issubclass(type_hint, parent_class):
            return _map_types_to_readable_names[parent_class]

    return None


def _get_component_from_input(hint, default_value=None):
    """
    Get FastComponent to represent the given input.

    This implementation returns the FastComponent associated with the specific input
    from the hint (type) of the input, type of the default value and the default value itself.

    Args:
        hint (FastComponent or Python data type): Either the input component or data type of the input.
        default_value (any, optional): Default value of the input component. Defaults to None.

    Returns:
        FastComponent: Component that can be used to represent the given input.
    """

    # If hint has the attribute "component_property", it indicates that hint is a FastComponent, return it
    if hasattr(hint, "component_property"):
        return hint

    # Elif the component is a Dash component, use the specified component_property
    # If one isn't specified, get the default component_property
    # If not even that assign it the component_property "value" and return the FastComponent.
    elif isinstance(type(hint), dash.development.base_component.ComponentMeta):
        default_component_property = _get_default_property(type(hint))

        return Fastify(component=hint, component_property=default_component_property)

    # If hint is not type, assume that the user specified an object. Change it to type
    if not isinstance(hint, type):
        hint = type(hint)

    _hint_type = _get_readable_names_from_parent_classes(hint)
    _default_value_type = _get_readable_names_from_parent_classes(type(default_value))

    # If the hint is a PIL Image
    if issubclass(hint, PIL.Image.Image):
        _hint_type = "Image"

    # If the default is a PIL Image
    if isinstance(default_value, PIL.Image.Image):
        _default_value_type = "Image"

    if _hint_type == "Text":
        if _default_value_type == "Text":
            component = Fastify(
                dmc.Textarea(
                    value=default_value,
                    autosize=True,
                    minRows=4,
                ),
                "value",
                tag=_hint_type,
            )

        elif _default_value_type == "Numeric":
            component = Fastify(
                dbc.Input(value=default_value, type="number"), "value", tag=_hint_type
            )

        elif _default_value_type == "Sequence":
            component = Fastify(
                dcc.Dropdown(options=default_value), "value", tag=_hint_type
            )

        elif _default_value_type == "Dictionary":
            component = Fastify(
                dbc.Input(value=str(default_value)), "value", tag=_hint_type
            )

        elif _default_value_type == "Boolean":
            component = Fastify(
                dbc.Input(value=str(default_value)), "value", tag=_hint_type
            )

        elif _default_value_type == "Date":
            component = Fastify(
                dbc.Input(value=str(default_value)), "value", tag=_hint_type
            )

        elif _default_value_type == "Timestamp":
            component = Fastify(
                dbc.Input(value=str(default_value)), "value", tag=_hint_type
            )

        else:
            warnings.warn("Unknown or unsupported default value type. Assuming text.")
            component = Text

    elif _hint_type == "Numeric":
        if _default_value_type == "Text":
            component = Fastify(
                dbc.Input(value=hint(default_value), type="number"),
                "value",
                tag=_hint_type,
            )

        elif _default_value_type == "Numeric":
            component = Fastify(
                dbc.Input(value=default_value, type="number"), "value", tag=_hint_type
            )

        elif _default_value_type == "Sequence":
            if isinstance(default_value, range):
                start, step, stop = (
                    default_value.start,
                    default_value.step,
                    default_value.stop,
                )

                component = Fastify(
                    dmc.Slider(min=start, max=stop, step=step), "value", tag=_hint_type
                )

            else:
                default_value = list(map(hint, default_value))
                start, stop = min(default_value), max(default_value)
                step = (stop - start) / len(default_value)

            component = Fastify(
                dcc.Slider(
                    start,
                    stop,
                    step,
                    marks=None,
                    tooltip={"placement": "bottom", "always_visible": True},
                ),
                "value",
                tag=_hint_type,
            )

        else:
            component = Fastify(dbc.Input(type="number"), "value", tag=_hint_type)

    elif _hint_type == "Sequence":
        if _default_value_type == "Text":
            component = Fastify(
                dmc.MultiSelect(
                    data=[default_value],
                    value=[default_value],
                    searchable=True,
                    creatable=True,
                ),
                "value",
                tag=_default_value_type,
            )

        elif _default_value_type == "Sequence":
            component = Fastify(
                dmc.MultiSelect(
                    data=default_value,
                    searchable=True,
                ),
                "value",
                tag=_default_value_type,
            )

        elif _default_value_type == "Dictionary":
            component = Fastify(
                dmc.MultiSelect(
                    data=default_value,
                    searchable=True,
                ),
                "value",
                tag=_default_value_type,
            )

        else:
            component = Fastify(
                dmc.MultiSelect(
                    searchable=True,
                    creatable=True,
                ),
                "value",
                tag=_default_value_type,
            )

    elif _hint_type == "Dictionary":
        component = Fastify(
            dcc.Dropdown(default_value, multi=True), "value", tag=_hint_type
        )

    elif _hint_type == "Boolean":
        if _default_value_type == "Boolean":
            component = Fastify(
                dbc.Checkbox(value=default_value), "value", tag=_hint_type
            )

        else:
            component = Fastify(dbc.Checkbox(), "value", tag=_hint_type)

    elif _hint_type == "Image":
        if _default_value_type == "Image":
            acknowledge_image_component = Fastify(
                component=html.Img(width="100%", style={"padding": "1% 0% 0% 0%"}),
                component_property="src",
                tag=_hint_type,
            )
            component = Fastify(
                component=dcc.Upload(
                    children=dbc.Col(["Click to upload image"]),
                    style={
                        "lineHeight": "60px",
                        "borderWidth": "1px",
                        "borderStyle": "dashed",
                        "borderRadius": "5px",
                        "textAlign": "center",
                    },
                    contents=_pil_to_b64(default_value),
                ),
                component_property="contents",
                ack=acknowledge_image_component,
                tag=_hint_type,
            )

        else:
            acknowledge_image_component = Fastify(
                component=html.Img(width="100%", style={"padding": "1% 0% 0% 0%"}),
                component_property="src",
            )
            component = Fastify(
                component=dcc.Upload(
                    children=dbc.Col(["Click to upload image"]),
                    style={
                        "lineHeight": "60px",
                        "borderWidth": "1px",
                        "borderStyle": "dashed",
                        "borderRadius": "5px",
                        "textAlign": "center",
                    },
                ),
                component_property="contents",
                ack=acknowledge_image_component,
                tag=_hint_type,
            )

    elif _hint_type == "Date":
        if _default_value_type == "Date":
            component = Fastify(
                dcc.DatePickerSingle(
                    display_format="MMM DD, YYYY",
                    date=default_value,
                    style={"width": "100%"},
                ),
                "date",
                tag=_hint_type,
            )

        else:
            component = Fastify(
                dcc.DatePickerSingle(
                    display_format="MMM DD, YYYY",
                    date=datetime.date.today(),
                    style={"width": "100%"},
                ),
                "date",
                tag=_hint_type,
            )

    else:
        if _default_value_type is not None:
            if _default_value_type == "Text":
                component = Fastify(
                    dbc.Input(value=default_value), "value", tag=_default_value_type
                )

            elif _default_value_type == "Numeric":
                component = Fastify(
                    dbc.Input(value=default_value, type="number"),
                    "value",
                    tag=_default_value_type,
                )

            elif _default_value_type == "Sequence":
                if isinstance(default_value, range):
                    start, step, stop = (
                        default_value.start,
                        default_value.step,
                        default_value.stop,
                    )
                    component = Fastify(
                        dcc.Slider(
                            start,
                            stop,
                            step,
                            marks=None,
                            tooltip={"placement": "bottom", "always_visible": True},
                        ),
                        "value",
                        tag=_default_value_type,
                    )

                else:
                    component = Fastify(
                        dcc.Dropdown(options=default_value),
                        "options",
                        tag=_default_value_type,
                    )

            elif _default_value_type == "Dictionary":
                component = Fastify(
                    dcc.Dropdown(options=default_value),
                    "value",
                    tag=_default_value_type,
                )

            elif _default_value_type == "Boolean":
                component = Fastify(
                    dbc.Checkbox(value=default_value), "value", tag=_default_value_type
                )

            elif _default_value_type == "Date":
                component = Fastify(
                    dcc.DatePickerSingle(
                        display_format="MMM DD, YYYY",
                        date=default_value,
                        style={"width": "100%"},
                    ),
                    "date",
                    tag=_default_value_type,
                )

            elif _default_value_type == "Timestamp":
                component = Fastify(
                    dcc.DatePickerSingle(
                        display_format="MMM DD, YYYY",
                        date=default_value.date(),
                        style={"width": "100%"},
                    ),
                    "date",
                    tag=_default_value_type,
                )

            elif _default_value_type == "Image":
                acknowledge_image_component = Fastify(
                    component=html.Img(width="100%", style={"padding": "1% 0% 0% 0%"}),
                    component_property="src",
                    tag=_default_value_type,
                )
                component = Fastify(
                    component=dcc.Upload(
                        children=dbc.Col(["Click to upload image"]),
                        style={
                            "lineHeight": "60px",
                            "borderWidth": "1px",
                            "borderStyle": "dashed",
                            "borderRadius": "5px",
                            "textAlign": "center",
                        },
                        contents=_pil_to_b64(default_value),
                    ),
                    component_property="contents",
                    ack=acknowledge_image_component,
                    tag=_default_value_type,
                )

            else:
                warnings.warn("Unknown or unsupported hint. Assuming text.")
                component = Text

        else:
            warnings.warn("Unknown or unsupported hint. Assuming text.")
            component = Text

    return component


def _get_output_components(_hint_type):
    if hasattr(_hint_type, "component_property"):
        return _hint_type

    # Elif the component is a Dash component, use the specified component_property
    # If one isn't specified, get the default component_property
    # If not even that assign it the component_property "value" and return the FastComponent.
    elif isinstance(type(_hint_type), dash.development.base_component.ComponentMeta):
        default_component_property = _get_default_property(type(_hint_type))

        return Fastify(
            component=_hint_type,
            component_property=default_component_property,
            tag=_hint_type,
        )

    # If hint is not type, assume that the user specified an object. Change it to type
    if not isinstance(_hint_type, type):
        _hint_type = type(_hint_type)

    if issubclass(_hint_type, PIL.Image.Image):
        component = Image

    elif issubclass(_hint_type, pd.DataFrame):
        component = Table

    elif _hint_type == mpl.figure.Figure:
        component = Image

    else:
        component = Fastify(html.H1(), "children", tag=_hint_type)

    return component


def _infer_input_components(func):
    signature = inspect.signature(func)
    components = []

    parameters = signature.parameters.items()
    for _, value in parameters:
        hint = value.annotation
        default = None if value.default == inspect._empty else value.default

        component = _get_component_from_input(hint, default)
        components.append(component)

    return components


def _infer_output_components(func, outputs, output_labels):
    signature = inspect.signature(func)
    components = []

    if isinstance(outputs, list):
        parameters = [(None, o) for o in outputs]

    elif outputs is not None:
        parameters = [(None, outputs)]

    else:
        parameters = list(
            enumerate(
                signature.return_annotation
                if isinstance(signature.return_annotation, tuple)
                else [signature.return_annotation]
            )
        )

    if output_labels is None:
        output_labels = [None] * len(parameters)

    if isinstance(output_labels, list) and len(output_labels) != len(parameters):
        raise ValueError(
            "Length of output labels must be equal to the number of outputs."
        )

    for (_, hint), label in zip(parameters, output_labels):
        component = _get_output_components(hint)
        component.label_ = label
        components.append(copy.deepcopy(component))

    return components


###################################
# Define default components #
###################################

##### General components
Text = Fastify(
    component=dmc.Textarea(
        placeholder="",
        autosize=True,
        minRows=4,
    ),
    component_property="value",
    placeholder="",
    tag="Text",
)

TextArea = Fastify(component=dbc.Textarea(), component_property="value", tag="Text")

Slider = Fastify(
    component=dcc.Slider(
        min=0,
        max=20,
        step=1,
        value=10,
        tooltip={"placement": "top", "always_visible": True},
    ),
    component_property="value",
    tag="Numeric",
)


##### Input components
Upload = Fastify(
    component=dcc.Upload(
        children=dbc.Col(["Click to upload"]),
        style={
            "lineHeight": "60px",
            "borderWidth": "1px",
            "borderStyle": "dashed",
            "borderRadius": "5px",
            "textAlign": "center",
        },
    ),
    component_property="contents",
    tag="Text",
)

acknowledge_image_component = Fastify(
    component=html.Img(width="100%", style={"padding": "1% 0% 0% 0%"}),
    component_property="src",
    tag="Image",
)

UploadImage = Fastify(
    component=dcc.Upload(
        children=dbc.Col(["Click to upload image"]),
        style={
            "lineHeight": "60px",
            "borderWidth": "1px",
            "borderStyle": "dashed",
            "borderRadius": "5px",
            "textAlign": "center",
        },
    ),
    component_property="contents",
    ack=acknowledge_image_component,
    tag="Image",
)


##### Output components
Image = Fastify(
    component=html.Img(
        style={
            "object-fit": "contain",
            "max-height": "90%",
            "max-width": "100%",
            "height": "auto",
        }
    ),
    component_property="src",
    tag="Image",
)

Graph = Fastify(
    component=dcc.Graph(style=dict(height="100%", width="100%")),
    component_property="figure",
    placeholder=(
        go.Figure()
        .update_yaxes(visible=False, showticklabels=False)
        .update_xaxes(visible=False, showticklabels=False)
    ),
    tag="Graph",
)

Chat = Fastify(
    html.Div(
        style={
            "overflow-y": "scroll",
            "overflow-x": "hidden",
            "display": "flex",
            "flex-direction": "column-reverse",
        }
    ),
    "children",
    tag="Chat",
)

Table = Fastify(
    dash.dash_table.DataTable(
        page_size=100,
        page_action="native",
        sort_action="native",
        style_header={
            "backgroundColor": "white",
            "fontWeight": "bold",
            "color": "black",
            "textAlign": "center",
            "border": "1px solid #f0f0f0",
            "fontFamily": '"News Cycle","Arial Narrow Bold",sans-serif',
        },
        style_cell={
            "backgroundColor": "white",
            "color": "black",
            "textAlign": "center",
            "border": "1px solid #f0f0f0",
            "fontFamily": '"News Cycle","Arial Narrow Bold",sans-serif',
        },
        style_table={
            "border": "1px solid #f0f0f0",
            "overflowY": "auto",
            "fontFamily": '"News Cycle","Arial Narrow Bold",sans-serif',
        },
    ),
    "data",
)



import functools
import logging
import re
import traceback
import warnings

import dash
import flask
from dash import Input, Output, ctx
from dash.exceptions import PreventUpdate

from .Components import (
    BaseLayout,
    SidebarLayout,
    _infer_input_components,
    _infer_output_components,
)

from .utils import (
    _assign_ids_to_inputs,
    _assign_ids_to_outputs,
    _make_input_groups,
    _make_output_groups,
    _transform_inputs,
    _transform_outputs,
    theme_mapper,
    _infer_variable_names,
    _parse_docstring_as_markdown,
    _get_error_notification_component,
)


class FastDash:
    """
    Fast Dash app object containing automatically generated UI components and callbacks.

    This is the primary Fast Dash data structure. Can be thought of as a wrapper around
    a flask WSGI application. It has in-built support for automated UI generation and
    sets all parameters required for Fast Dash app deployment.
    """

    def __init__(
        self,
        callback_fn,
        layout="sidebar",
        mosaic=None,
        inputs=None,
        outputs=None,
        output_labels="infer",
        title=None,
        title_image_path=None,
        subheader=None,
        github_url=None,
        linkedin_url=None,
        twitter_url=None,
        navbar=True,
        footer=True,
        about=True,
        theme=None,
        update_live=False,
        port=8080,
        mode=None,
        minimal=False,
        disable_logs=False,
        scale_height=1,
        run_kwargs=dict(),
        **kwargs
    ):
        """
        Args:
            callback_fn (func): Python function that Fast Dash deploys. \
                This function guides the behavior of and interaction between input and output components.

            layout (str, optional): App layout style. Currently supports 'base' and 'sidebar'. Defaults to sidebar.

            mosaic (str): Mosaic array layout, if sidebar layout is selected.

            inputs (Fast component, list of Fast components, optional): Components to represent inputs of the callback function.\
                Defaults to None. If `None`, Fast Dash attempts to infer the best components from callback function's type \
                hints and default values. In the absence of type hints, default components are all `Text`.

            outputs (Fast component, list of Fast components, optional): Components to represent outputs of the callback function.\
                Defaults to None. If `None`, Fast Dash attempts to infer the best components from callback function's type hints.\
                In the absence of type hints, default components are all `Text`.

            output_labels(list of string labels or "infer" or None, optional): Labels given to the output components. If None, inputs are\
                set labeled integers starting at 1 (Output 1, Output 2, and so on). If "infer", labels are inferred from the function\
                signature. Defaults to infer.

            title (str, optional): Title given to the app. If `None`, function name (assumed to be in snake case)\
                is converted to title case. Defaults to None.

            title_image_path (str, optional): Path (local or URL) of the app title image. Defaults to None.

            subheader (str, optional): Subheader of the app, displayed below the title image and title\
                If `None`, Fast Dash tries to use the callback function's docstring instead. Defaults to None.
                

            github_url (str, optional): GitHub URL for branding. Displays a GitHub logo in the navbar, which takes users to the\
                specified URL. Defaults to None.

            linkedin_url (str, optional): LinkedIn URL for branding Displays a LinkedIn logo in the navbar, which takes users to the\
                specified URL. Defaults to None.

            twitter_url (str, optional): Twitter URL for branding. Displays a Twitter logo in the navbar, which takes users to the\
                specified URL. Defaults to None.

            navbar (bool, optional): Display navbar. Defaults to True.

            footer (bool, optional): Display footer. Defaults to True.

            about (Union[str, bool], optional): App description to display on clicking the `About` button. If True, content is inferred from\
                the docstring of the callback function. If string, content is used directly as markdown. \
                `About` is hidden if False or None. Defaults to True.

            theme (str, optional): Apply theme to the app.All available themes can be found at https://bootswatch.com/. Defaults to JOURNAL. 

            update_live (bool, optional): Enable hot reloading. If the number of inputs is 0, this is set to True automatically. Defaults to False.

            port (int, optional): Port to which the app should be deployed. Defaults to 8080.

            mode (str, optional): Mode in which to launch the app. Acceptable options are `None`, `jupyterlab`, `inline`, 'external`.\
                Defaults to None.

            minimal (bool, optional): Display minimal version by hiding navbar, title, title image, subheader and footer. Defaults to False.

            disable_logs (bool, optional): Hide app logs. Sets logger level to `ERROR`. Defaults to False.

            scale_height (float, optional): Height of the app container is enlarged as a multiple of this. Defaults to 1.

            run_kwargs (dict, optional): All values from this variable are passed to Dash's `.run` method.
        """

        self.callback_fn = callback_fn
        self.layout_pattern = layout
        self.mosaic = mosaic
        self.output_labels = output_labels

        if output_labels == "infer":
            self.output_labels = _infer_variable_names(callback_fn)

        self.inputs = (
            _infer_input_components(callback_fn)
            if inputs is None
            else inputs
            if isinstance(inputs, list)
            else [inputs]
        )
        self.outputs = _infer_output_components(
            callback_fn, outputs, self.output_labels
        )
        self.update_live = (
            True
            if (isinstance(self.inputs, list) and len(self.inputs) == 0)
            else update_live
        )
        self.mode = mode
        self.disable_logs = disable_logs
        self.scale_height = scale_height
        self.port = port
        self.run_kwargs = run_kwargs
        self.run_kwargs.update(dict(port=port))
        self.kwargs = kwargs

        if self.disable_logs is True:
            log = logging.getLogger("werkzeug")
            log.setLevel(logging.ERROR)

        else:
            log = logging.getLogger("werkzeug")
            log.setLevel(logging.DEBUG)

        if title is None:
            title = re.sub("[^0-9a-zA-Z]+", " ", callback_fn.__name__).title()

        self.title = title
        self.title_image_path = title_image_path
        self.subtitle = (
            subheader
            if subheader is not None
            else _parse_docstring_as_markdown(
                callback_fn, title=self.title, get_short=True
            )
        )
        self.github_url = github_url
        self.linkedin_url = linkedin_url
        self.twitter_url = twitter_url
        self.navbar = navbar
        self.footer = footer
        self.about = about
        self.theme = theme or "JOURNAL"
        self.minimal = minimal

        # Extract input tags
        self.input_tags = [inp.tag for inp in self.inputs]
        self.output_tags = [inp.tag for inp in self.outputs]

        # Assign IDs to components
        self.inputs_with_ids = _assign_ids_to_inputs(self.inputs, self.callback_fn)
        self.outputs_with_ids = _assign_ids_to_outputs(self.outputs)
        self.ack_mask = [
            False if (not hasattr(input_, "ack") or (input_.ack is None)) else True
            for input_ in self.inputs_with_ids
        ]

        # Default state of outputs
        self.output_state_default = [
            output_.placeholder if hasattr(output_, "placeholder") else None
            for output_ in self.outputs_with_ids
        ]

        self.output_state_blank = [None for output_ in self.outputs_with_ids]
        self.latest_output_state = self.output_state_blank

        # Define Flask server
        server = flask.Flask(__name__)
        external_stylesheets = [
            theme_mapper(self.theme),
            "https://use.fontawesome.com/releases/v5.9.0/css/all.css",
        ]

        source = dash.Dash
        # if self.mode is not None:
        #     try:
        #         from jupyter_dash import JupyterDash

        #         source = JupyterDash

        #     except ImportError as e:
        #         self.mode = None
        #         warnings.warn(str(e))
        #         warnings.warn("Ignoring mode argument")

        self.app = source(
            __name__,
            external_stylesheets=external_stylesheets,
            server=server,
            **self.kwargs
        )
        # Define app title
        self.app.title = self.title or ""

        # Intialize layout
        self.set_layout()

        # Register callbacks
        self.register_callback_fn()

        # Keep track of the number of clicks
        self.submit_clicks = 0
        self.reset_clicks = 0
        self.app_initialized = False

        # Allow easier access to Dash server
        self.server = self.app.server

    def run(self):
        self.app.run(**self.run_kwargs) if self.mode is None else self.app.run_server(
            jupyter_mode=self.mode, **self.run_kwargs
        )

    def run_server(self):
        self.app.run_server(
            **self.run_kwargs
        ) if self.mode is None else self.app.run_server(
            jupyter_mode=self.mode, **self.run_kwargs
        )

    def set_layout(self):
        if self.inputs is not None:
            input_groups = _make_input_groups(self.inputs_with_ids, self.update_live)

        if self.outputs is not None:
            output_groups = _make_output_groups(self.outputs_with_ids, self.update_live)

        layout_args = {
            "mosaic": self.mosaic,
            "inputs": input_groups,
            "outputs": output_groups,
            "title": self.title,
            "title_image_path": self.title_image_path,
            "subtitle": self.subtitle,
            "github_url": self.github_url,
            "linkedin_url": self.linkedin_url,
            "twitter_url": self.twitter_url,
            "navbar": self.navbar,
            "footer": self.footer,
            "about": self.about,
            "minimal": self.minimal,
            "scale_height": self.scale_height,
            "app": self,
        }

        if self.layout_pattern == "sidebar":
            app_layout = SidebarLayout(**layout_args)

        else:
            app_layout = BaseLayout(**layout_args)

        self.layout_object = app_layout
        self.app.layout = app_layout.generate_layout()

    def register_callback_fn(self):
        @self.app.callback(
            [
                Output(
                    component_id=output_.id,
                    component_property=output_.component_property,
                )
                for output_ in self.outputs_with_ids
            ]
            + [Output("error-notify-div", "children")],
            [
                Input(
                    component_id=input_.id, component_property=input_.component_property
                )
                for input_ in self.inputs_with_ids
            ]
            + [
                Input(component_id="reset_inputs", component_property="n_clicks"),
                Input(component_id="submit_inputs", component_property="n_clicks"),
            ],
            prevent_initial_callback=True,
        )
        def process_input(*args):
            if (
                ctx.triggered_id not in ["submit_inputs", "reset_inputs"]
                and self.update_live is False
            ):
                raise PreventUpdate

            default_notification = None

            try:
                inputs = _transform_inputs(args[:-2], self.input_tags)

                if ctx.triggered_id == "submit_inputs" or (
                    self.update_live is True and None not in args
                ):
                    self.app_initialized = True

                    output_state = self.callback_fn(*inputs)

                    if isinstance(output_state, tuple):
                        self.output_state = list(output_state)

                    else:
                        self.output_state = [output_state]

                    # Transform outputs to fit in the desired components
                    self.output_state = _transform_outputs(
                        self.output_state, self.output_tags
                    )

                    # Log the latest output state
                    self.latest_output_state = self.output_state

                    return self.output_state + [default_notification]

                elif ctx.triggered_id == "reset_inputs":
                    self.output_state = self.output_state_default
                    return self.output_state + [default_notification]

                elif self.app_initialized:
                    return self.output_state + [default_notification]

                else:
                    return self.output_state_default + [default_notification]

            except Exception as e:
                traceback.print_exc()
                notification = _get_error_notification_component(str(e))

                return self.output_state_default + [notification]

        @self.app.callback(
            [
                Output(
                    component_id=input_.ack.id,
                    component_property=input_.ack.component_property,
                )
                for input_ in self.inputs_with_ids
            ]
            + [Output("dummy-div", "children")],
            [
                Input(
                    component_id=input_.id, component_property=input_.component_property
                )
                for input_ in self.inputs_with_ids
            ]
            + [Input("dummy-div", "children")],
        )
        def process_ack_outputs(*args):
            ack_components = [
                ack if mask is True else None
                for mask, ack in zip(self.ack_mask, list(args)[:-1])
            ]
            return ack_components + [[]]

        # Set layout callbacks
        if not self.minimal:
            self.layout_object.callbacks(self)


def fastdash(
    _callback_fn=None,
    *,
    layout="sidebar",
    mosaic=None,
    inputs=None,
    outputs=None,
    output_labels="infer",
    title=None,
    title_image_path=None,
    subheader=None,
    github_url=None,
    linkedin_url=None,
    twitter_url=None,
    navbar=True,
    footer=True,
    about=True,
    theme=None,
    update_live=False,
    port=8080,
    mode=None,
    minimal=False,
    disable_logs=False,
    scale_height=1,
    run_kwargs=dict(),
    **kwargs
):
    """
    Function decorator / wrapper for Fast Dash.

    Args:
        callback_fn (func): Python function that Fast Dash deploys. \
            This function guides the behavior of and interaction between input and output components.

        layout (str, optional): App layout style. Currently supports 'base' and 'sidebar'. Defaults to sidebar.

        mosaic (str): Mosaic array layout, if sidebar layout is selected.

        inputs (Fast component, list of Fast components, optional): Components to represent inputs of the callback function.\
            Defaults to None. If `None`, Fast Dash attempts to infer the best components from callback function's type \
            hints and default values. In the absence of type hints, default components are all `Text`.

        outputs (Fast component, list of Fast components, optional): Components to represent outputs of the callback function.\
            Defaults to None. If `None`, Fast Dash attempts to infer the best components from callback function's type hints.\
            In the absence of type hints, default components are all `Text`.

        output_labels(list of string labels or "infer" or None, optional): Labels given to the output components. If None, inputs are\
            set labeled integers starting at 1 (Output 1, Output 2, and so on). If "infer", labels are inferred from the function\
            signature. Defaults to infer.

        title (str, optional): Title given to the app. If `None`, function name (assumed to be in snake case)\
            is converted to title case. Defaults to None.

        title_image_path (str, optional): Path (local or URL) of the app title image. Defaults to None.

        subheader (str, optional): Subheader of the app, displayed below the title image and title\
            If `None`, Fast Dash tries to use the callback function's docstring instead. Defaults to None.
            

        github_url (str, optional): GitHub URL for branding. Displays a GitHub logo in the navbar, which takes users to the\
            specified URL. Defaults to None.

        linkedin_url (str, optional): LinkedIn URL for branding Displays a LinkedIn logo in the navbar, which takes users to the\
            specified URL. Defaults to None.

        twitter_url (str, optional): Twitter URL for branding. Displays a Twitter logo in the navbar, which takes users to the\
            specified URL. Defaults to None.

        navbar (bool, optional): Display navbar. Defaults to True.

        footer (bool, optional): Display footer. Defaults to True.

        about (Union[str, bool], optional): App description to display on clicking the `About` button. If True, content is inferred from\
            the docstring of the callback function. If string, content is used directly as markdown. \
            `About` is hidden if False or None. Defaults to True.

        theme (str, optional): Apply theme to the app.All available themes can be found at https://bootswatch.com/. Defaults to JOURNAL. 

        update_live (bool, optional): Enable hot reloading. If the number of inputs is 0, this is set to True automatically. Defaults to False.

        port (int, optional): Port to which the app should be deployed. Defaults to 8080.

        mode (str, optional): Mode in which to launch the app. Acceptable options are `None`, `jupyterlab`, `inline`, 'external`.\
            Defaults to None.

        minimal (bool, optional): Display minimal version by hiding navbar, title, title image, subheader and footer. Defaults to False.

        disable_logs (bool, optional): Hide app logs. Sets logger level to `ERROR`. Defaults to False.

        scale_height (float, optional): Height of the app container is enlarged as a multiple of this. Defaults to 1.

        run_kwargs (dict, optional): All values from this variable are passed to Dash's `.run` method.
        """

    def decorator_fastdash(callback_fn):
        "Decorator for callback_fn"

        @functools.wraps(callback_fn)
        def wrapper_fastdash(**kwargs):
            app = FastDash(callback_fn=callback_fn, **kwargs)
            app.run()
            return callback_fn

        return wrapper_fastdash(
            layout=layout,
            mosaic=mosaic,
            inputs=inputs,
            outputs=outputs,
            output_labels=output_labels,
            title=title,
            title_image_path=title_image_path,
            subheader=subheader,
            github_url=github_url,
            linkedin_url=linkedin_url,
            twitter_url=twitter_url,
            navbar=navbar,
            footer=footer,
            about=about,
            theme=theme,
            update_live=update_live,
            mode=mode,
            port=port,
            minimal=minimal,
            disable_logs=disable_logs,
            scale_height=scale_height,
            run_kwargs=run_kwargs,
            **kwargs
        )

    # If the decorator is called with arguments
    if _callback_fn is None:
        return decorator_fastdash
    # If the decorator is called without arguments. Use default input and output values
    else:
        return decorator_fastdash(_callback_fn)



#!/usr/bin/env python
"""Tests for `fast_dash` package."""
# pylint: disable=redefined-outer-name

from fast_dash import FastDash
from fast_dash.Components import Text
import matplotlib.pyplot as plt

import time

import pytest
from selenium.common.exceptions import NoSuchElementException


## Define callback functions
def simple_text_to_text_function(input_text):
    "Converts text to text"
    return input_text


def simple_text_to_multiple_text_function(input_text):
    return input_text, input_text


def simple_text_to_multiple_outputs(
    input_text: str,
) -> (plt.Figure, str):
    "Something"

    fig, ax = plt.subplots(1, 1)
    ax.plot([1, 2, 3], [1, 2, 3])

    return fig, "Return some text"


def test_fdfd001_set_title(dash_duo):
    "Test title element"
    app = FastDash(
        callback_fn=simple_text_to_text_function,
        inputs=Text,
        outputs=Text,
        title="App title",
    ).app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal("#title8888928", "App title", timeout=4)

    assert dash_duo.find_element("#title8888928").text == "App title"
    assert dash_duo.get_logs() in [[], None], "browser console should contain no error"


def test_fdfd002_set_default_title(dash_duo):
    "Test default title"
    app = FastDash(
        callback_fn=simple_text_to_text_function, inputs=Text, outputs=Text
    ).app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal(
        "#title8888928", "Simple Text To Text Function", timeout=4
    )

    assert dash_duo.find_element("#title8888928").text == "Simple Text To Text Function"
    assert dash_duo.get_logs() in [[], None], "browser console should contain no error"


def test_fdfd003_output_is_none(dash_duo):
    "Test if the output is set to Text if none is specified"
    app = FastDash(callback_fn=simple_text_to_text_function, inputs=Text).app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal(
        "#title8888928", "Simple Text To Text Function", timeout=4
    )

    assert dash_duo.get_logs() in [[], None], "browser console should contain no error"


def test_fdfd004_click_submit(dash_duo):
    "Test clicking the submit button"
    app = FastDash(callback_fn=simple_text_to_text_function, inputs=Text).app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal(
        "#title8888928", "Simple Text To Text Function", timeout=4
    )

    # Enter some text
    form_textfield = dash_duo.find_element("#input_text")
    form_textfield.send_keys("Sample text")

    # Click submit
    dash_duo.multiple_click("#submit_inputs", 1)
    dash_duo.wait_for_text_to_equal("#output-1", "Sample text", timeout=4)

    # Click clear
    dash_duo.multiple_click("#reset_inputs", 1)
    dash_duo.wait_for_text_to_equal("#output-1", "", timeout=4)


def test_fdfd005_multiple_outputs(dash_duo):
    "Test what happens when a function returns a tuple as the output"
    app = FastDash(
        callback_fn=simple_text_to_multiple_text_function,
        inputs=Text,
        outputs=[Text, Text],
    ).app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal(
        "#title8888928", "Simple Text To Multiple Text Function", timeout=4
    )

    # Enter some text
    form_textfield = dash_duo.find_element("#input_text")
    form_textfield.send_keys("Sample text")

    # Click submit
    dash_duo.multiple_click("#submit_inputs", 1)
    dash_duo.wait_for_text_to_equal("#output-1", "Sample text", timeout=4)
    dash_duo.wait_for_text_to_equal("#output-2", "Sample text", timeout=4)

    # Click clear
    dash_duo.multiple_click("#reset_inputs", 1)
    dash_duo.wait_for_text_to_equal("#output-1", "", timeout=4)
    dash_duo.wait_for_text_to_equal("#output-2", "", timeout=4)


def test_fdfd006_live_update(dash_duo):
    "Test live update functionality"
    app = FastDash(
        callback_fn=simple_text_to_text_function,
        inputs=Text,
        outputs=Text,
        title="App title",
        update_live=True,
    ).app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal("#title8888928", "App title", timeout=4)

    assert dash_duo.find_element("#title8888928").text == "App title"
    assert dash_duo.get_logs() in [[], None], "browser console should contain no error"

    dash_duo.percy_snapshot("fdfd006-layout")


def test_fdfd007_subheader_docstring(dash_duo):
    "Test if subheader is function doc string if set to None"
    app = FastDash(
        callback_fn=simple_text_to_text_function,
        inputs=Text,
        outputs=Text,
        title="App title",
        update_live=True,
    ).app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal("#title8888928", "App title", timeout=4)

    assert dash_duo.find_element("#subheader6904007").text == "Converts text to text"
    assert dash_duo.get_logs() in [[], None], "browser console should contain no error"

    dash_duo.percy_snapshot("fdfd007-layout")


def test_fdfd008_minimal_mode(dash_duo):
    "Test Fast Dash's minimal mode"
    app = FastDash(
        callback_fn=simple_text_to_text_function,
        inputs=Text,
        outputs=Text,
        minimal=True,
    ).app

    dash_duo.start_server(app)
    time.sleep(4)

    with pytest.raises(NoSuchElementException):
        dash_duo.find_element("title8888928")

    with pytest.raises(NoSuchElementException):
        dash_duo.find_element("subheader6904007")

    with pytest.raises(NoSuchElementException):
        dash_duo.find_element("navbar3260780")

    with pytest.raises(NoSuchElementException):
        dash_duo.find_element("header1162572")

    with pytest.raises(NoSuchElementException):
        dash_duo.find_element("footer5265971")

    assert dash_duo.get_logs() in [[], None], "browser console should contain no error"

    dash_duo.percy_snapshot("fdfd008-layout")


def test_fdfd009_jupyter_dash(dash_duo):
    "Test Fast Dash's mode argument"

    app = FastDash(
        callback_fn=simple_text_to_text_function,
        inputs=Text,
        mode="inline",
    ).app

    dash_duo.start_server(app)
    time.sleep(4)


def test_fdfd010_output_labels(dash_duo):
    "Test the auto-infer variable name feature"

    app = FastDash(callback_fn=simple_text_to_multiple_outputs)
    assert app.output_labels == ["FIG", "RETURN_SOME_TEXT"]


def test_fdfd011_base_layout(dash_duo):
    "Test base layout"

    app = FastDash(callback_fn=simple_text_to_multiple_outputs, layout="base")
    assert app.output_labels == ["FIG", "RETURN_SOME_TEXT"]


def test_fdfd012_about_button_true(dash_duo):
    "Test the about button auto-documentation generation"

    def example_function(param1, param2=42):
        """
        An example function for demonstration.

        Args:
            param1 (str): Description for parameter 1.
            param2 (int, optional): Description for parameter 2. Defaults to 42.

        Returns:
            bool: Description for return value.
        """
        return True

    app = FastDash(callback_fn=example_function, inputs=Text, outputs=Text).app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal("#title8888928", "Example Function", timeout=4)

    # Click About
    dash_duo.multiple_click("#about-navlink", 1)
    time.sleep(2)

    # Grab the generated markdown text
    displayed_markdown = dash_duo.find_element("#about-modal-body").text

    assert "Example Function" in displayed_markdown, "Docstring absent in about (1)"
    assert (
        "Description for parameter 1." in displayed_markdown
    ), "Docstring absent in about (2)"


def test_fdfd013_about_button_false(dash_duo):
    "Test if about button is False"

    def example_function(param1, param2=42):
        """
        An example function for demonstration.

        Args:
            param1 (str): Description for parameter 1.
            param2 (int, optional): Description for parameter 2. Defaults to 42.

        Returns:
            bool: Description for return value.
        """
        return True

    app = FastDash(
        callback_fn=example_function, inputs=Text, outputs=Text, about=False
    ).app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal("#title8888928", "Example Function", timeout=4)

    # Click About
    assert not dash_duo.find_elements(f"#about-navlink")


def test_fdfd014_about_button_custom(dash_duo):
    "Test the about button custom documentation"

    def example_function(param1, param2=42):
        """
        An example function for demonstration.

        Args:
            param1 (str): Description for parameter 1.
            param2 (int, optional): Description for parameter 2. Defaults to 42.

        Returns:
            bool: Description for return value.
        """
        return True

    # When about argument is custom text
    app = FastDash(
        callback_fn=example_function,
        inputs=Text,
        outputs=Text,
        about="Custom about section",
    ).app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal("#title8888928", "Example Function", timeout=4)

    # Click About
    dash_duo.multiple_click("#about-navlink", 1)
    time.sleep(2)

    # Grab the generated markdown text
    displayed_markdown = dash_duo.find_element("#about-modal-body").text

    assert (
        "Custom about section" in displayed_markdown
    ), "Docstring mismatch in about (3)"


def test_fdfd015_close_sidebar(dash_duo):
    "Test closing the sidebar"

    app = FastDash(
        callback_fn=simple_text_to_text_function, inputs=Text, outputs=Text
    ).app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal(
        "#title8888928", "Simple Text To Text Function", timeout=4
    )

    # Click sidebar toggle
    dash_duo.multiple_click("#sidebar-button", 1)
    time.sleep(2)

    # Find the style of the sidebar
    sidebar_style = dash_duo.find_element("#input-group").get_attribute("style")
    sidebar_style = dict(
        item.split(":") for item in sidebar_style.strip(";").split("; ") if item
    )
    assert sidebar_style["display"].strip() == "none", "Sidebar did not close"



#!/usr/bin/env python
"""Tests for `fast_dash` package."""
# pylint: disable=redefined-outer-name

from .examples import (
    example_1_simple_text_to_text,
    example_2_text_with_slider,
    example_3_image_to_image,
    example_4_image_slider_to_image_text,
    example_5_uploadimage_to_image,
    example_6_text_to_plt
)


def test_example_1(dash_duo):
    "Test example_1_simple_text_to_text"

    app = example_1_simple_text_to_text().app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal("#title8888928", "Fast Dash example 1", timeout=4)

    assert dash_duo.find_element("#title8888928").text == "Fast Dash example 1"
    assert dash_duo.get_logs() in [[], None], "browser console should contain no error"


def test_example_2(dash_duo):
    "Test example_2_text_with_slider"

    app = example_2_text_with_slider().app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal("#title8888928", "Fast Dash example 2", timeout=4)

    assert dash_duo.find_element("#title8888928").text == "Fast Dash example 2"
    assert dash_duo.get_logs() in [[], None], "browser console should contain no error"


def test_example_3(dash_duo):
    "Test example_3_image_to_image"

    app = example_3_image_to_image().app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal("#title8888928", "Fast Dash example 3", timeout=4)

    assert dash_duo.find_element("#title8888928").text == "Fast Dash example 3"
    assert dash_duo.get_logs() in [[], None], "browser console should contain no error"


def test_example_4(dash_duo):
    "Test example_4_image_slider_to_image_text"

    app = example_4_image_slider_to_image_text().app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal("#title8888928", "Fast Dash example 4", timeout=4)

    assert dash_duo.find_element("#title8888928").text == "Fast Dash example 4"
    assert dash_duo.get_logs() in [[], None], "browser console should contain no error"


def test_example_5(dash_duo):
    "Test example_5_uploadimage_to_image"

    app = example_5_uploadimage_to_image().app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal("#title8888928", "Fast Dash example 5", timeout=4)

    assert dash_duo.find_element("#title8888928").text == "Fast Dash example 5"
    assert dash_duo.get_logs() in [[], None], "browser console should contain no error"

def test_example_6(dash_duo):
    "Test example_6_text_to_plt"

    app = example_6_text_to_plt().app

    dash_duo.start_server(app)
    dash_duo.wait_for_text_to_equal("#title8888928", "Fast Dash example 6", timeout=4)

    assert dash_duo.find_element("#title8888928").text == "Fast Dash example 6"
    assert dash_duo.get_logs() in [[], None], "browser console should contain no error"



"""
Utility functions
"""
import base64
import copy
import inspect
from io import BytesIO

from dash import html, dcc, Patch
import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
from dash_iconify import DashIconify

import docstring_parser
import logging
import matplotlib as mpl
import matplotlib.pyplot as plt
import PIL
import pandas as pd
import warnings
import re


def Fastify(component, component_property, ack=None, placeholder=None, label_=None,tag=None, *args, **kwargs):
    """
    Modify a Dash component into a FastComponent.

    Args:
        component (Dash component): Dash component that needs to be modified
        component_property (str): Component property that's assigned the input or output values
        ack (Dash component, optional): Dash component that's displayed as an acknowledgement of the original component
        placeholder (str, optional): Placeholder value of the component.
        label_ (str, optional): Component title.
        tag (str, optional): Optional tag applied to the component.

    Returns:
        Fast component: Dash component modified to make it compatible with Fast Dash.
    """
    
    class FastComponent(type(component)):
        def __init__(self, component, component_property, ack=ack, placeholder=placeholder, label_=label_, tag=tag, *args, **kwargs):
            
            self.component_property = component_property
            self.ack = ack
            self.label_ = label_
            self.placeholder = placeholder
            self.tag = tag
            
            # Copy normal attributes
            for attr_name, attr_value in vars(component).items():
                setattr(self, attr_name, attr_value)

            # Copy the __doc__ attribute
            self.__doc__ = component.__doc__
            
            for key, value in kwargs.items():
                setattr(self, key, value)
                
        def __call__(self, **kwargs):            
            self_copy = copy.deepcopy(self)
            for key, value in kwargs.items():
                setattr(self_copy, key, value)
                
            return self_copy
        
    return FastComponent(component, component_property, ack=ack, placeholder=placeholder, label_=label_, tag=tag)


def Chatify(query_response_dict):
    "Convert a dictionary into a Chat component"

    if not isinstance(query_response_dict, dict):
        raise TypeError("Chat component requires a dictionary output ('query': ..., 'response': ...).")

    ## Chat component
    input_component = dbc.Row(
        dmc.Text(
            [query_response_dict["query"]],
            align="end",
            style={
                "padding": "1% 1%",
                "max-width": "80%",
                "backgroundColor": "#E8EBFA",
            },
            className="border rounded shadow-sm m-3 col-auto",
        ),
        align="center",
        justify="end",
    )

    output_component = dbc.Row(
        [
            dmc.Text(
                [
                    dbc.Col(
                        DashIconify(
                            icon="ic:baseline-question-answer",
                            color="#910517",
                            width=30,
                        ),
                        class_name="pb-2",
                    ),
                    dcc.Markdown(query_response_dict["response"]),
                ],
                align="start",
                style={
                    "padding": "1% 1%",
                    "max-width": "98%",
                    "backgroundColor": "#F9F9F9",
                },
                className="border rounded shadow-sm m-3",
            )
        ],
        align="start",
        justify="start",
    )

    chat_output = Patch()
    chat_output.prepend(input_component)
    chat_output.prepend(output_component)

    return chat_output


# Add themes mapper
def theme_mapper(theme_name):
    """
    Map theme name string ot a dbc theme object
    """

    theme_mapper_dict = {
        "BOOTSTRAP": dbc.themes.BOOTSTRAP,
        "CERULEAN": dbc.themes.CERULEAN,
        "COSMO": dbc.themes.COSMO,
        "CYBORG": dbc.themes.CYBORG,
        "DARKLY": dbc.themes.DARKLY,
        "FLATLY": dbc.themes.FLATLY,
        "JOURNAL": dbc.themes.JOURNAL,
        "LITERA": dbc.themes.LITERA,
        "LUMEN": dbc.themes.LUMEN,
        "LUX": dbc.themes.LUX,
        "MATERIA": dbc.themes.MATERIA,
        "MINTY": dbc.themes.MINTY,
        "MORPH": dbc.themes.MORPH,
        "PULSE": dbc.themes.PULSE,
        "QUARTZ": dbc.themes.QUARTZ,
        "SANDSTONE": dbc.themes.SANDSTONE,
        "SIMPLEX": dbc.themes.SIMPLEX,
        "SKETCHY": dbc.themes.SKETCHY,
        "SLATE": dbc.themes.SLATE,
        "SOLAR": dbc.themes.SOLAR,
        "SPACELAB": dbc.themes.SPACELAB,
        "SUPERHERO": dbc.themes.SUPERHERO,
        "UNITED": dbc.themes.UNITED,
        "VAPOR": dbc.themes.VAPOR,
        "YETI": dbc.themes.YETI,
        "ZEPHYR": dbc.themes.ZEPHYR,
    }

    theme = theme_mapper_dict.get(theme_name.upper(), dbc.themes.YETI)

    return theme


# Data type conversion utilities
def _pil_to_b64(img):
    """
    Utility to convert PIL image to a base64 string.

    Args:
        img (PIL.Image): Input image

    Returns:
        str: Base64 string of the input image
    """
    buffered = BytesIO()
    img.save(buffered, format="PNG")
    img_str = base64.b64encode(buffered.getvalue()).decode("utf-8")
    img_str = f"data:image/png;base64,{img_str}"

    return img_str


def _b64_to_pil(img_str):
    """
    Utility to convert a base64 image  string to PIL image.

    Args:
        img_str (str): Input image base64 string

    Returns:
        PIL.Image: Pillow image formatted image
    """

    img_str = img_str.split(";base64,")[1]
    return PIL.Image.open(BytesIO(base64.b64decode(img_str)))


def _mpl_to_b64(fig):
    """
    Utility to convert Matplotlib figure to a base64 string.

    Args:
        fig (matplotlib.figure.Figure): Input matplotlib plot

    Returns:
        str: Base64 string of the input image
    """
    buffered = BytesIO()
    fig.savefig(buffered, format="PNG")
    img_str = base64.b64encode(buffered.getvalue()).decode("utf-8")
    img_str = f"data:image/png;base64,{img_str}"

    return img_str


# Input component utils
def _get_input_names_from_callback_fn(callback_fn):
    """
    Returns the names of function arguments as a list of strings
    """
    signature = inspect.signature(callback_fn)
    parameters = signature.parameters
    parameter_list = list(parameters)

    return parameter_list


# Fast Dash app utilities
def _assign_ids_to_inputs(inputs, callback_fn):
    """
    Modify the 'id' property of inputs.
    """
    if not isinstance(inputs, list):
        inputs = [inputs]

    inputs_with_ids = []

    for input_, parameter_name in zip(
        inputs, _get_input_names_from_callback_fn(callback_fn)
    ):
        input_.id = parameter_name
        inputs_with_ids.append(copy.deepcopy(input_))

    return inputs_with_ids


def _make_input_groups(inputs_with_ids, update_live):
    input_groups = []

    input_groups.append(html.H4("INPUTS"))

    for idx, input_ in enumerate(inputs_with_ids):
        label = (
            f"{input_.id}"
            if (not hasattr(input_, "label_") or input_.label_ is None)
            else input_.label_
        )
        label = label.replace("_", " ").upper()
        ack_component = (
            Fastify(component=dbc.Col(), component_property="children")
            if (not hasattr(input_, "ack") or input_.ack is None)
            else input_.ack
        )
        ack_component.id = f"{input_.id}_ack"

        input_.ack = ack_component

        input_groups.append(
            dbc.Col(
                [dbc.Label(label, align="end"), input_, ack_component],
                align="left",
            )
        )

    button_row = html.Div(
        [
            dbc.Button(
                "Submit",
                color="warning",
                className="me-1",
                id="submit_inputs",
                n_clicks=0,
            )
        ],
        style={"padding": "2% 1% 1% 2%"}
        if update_live is False
        else dict(display="none"),
    )

    input_groups.append(button_row)

    return input_groups


# Output utils
def _assign_ids_to_outputs(outputs):
    """
    Modify the 'id' property of inputs.
    """
    if not isinstance(outputs, list):
        outputs = [outputs]

    outputs_with_ids = []

    for idx, output_ in enumerate(outputs):
        output_.id = f"output-{idx + 1}"
        outputs_with_ids.append(copy.deepcopy(output_))

    return outputs_with_ids


def _make_output_groups(outputs, update_live):
    output_groups = []
    # output_groups.append(html.H6("OUTPUT"))

    for idx, output_ in enumerate(outputs):
        label = f"Output {idx + 1}" if output_.label_ is None else output_.label_
        label = label.replace("_", " ")
        output_groups.append(
            dbc.Col(
                [dbc.Label(label, align="end")] + [output_],
                align="center",
                style={"width": "100%", "overflow": "hidden"},
                class_name="rounded border d-flex flex-column flex-fill",
            )
        )

    button_row = html.Div(
        [
            dbc.Button(
                "Clear",
                outline=True,
                color="primary",
                className="me-1",
                id="reset_inputs",
                n_clicks=0,
            )
        ],
        style={"padding": "2% 1% 1% 2%"}
        if update_live is False
        else dict(display="none"),
    )

    output_groups.append(button_row)

    return output_groups


def _get_transform_function(output, tag):
    "Utility for _transform_outputs. Defines the transform function to be applied to a Fast component's property."

    if tag == "Chat":
        return Chatify

    subclass = type(output)
    _transform_mapper = {plt.Figure: _mpl_to_b64, 
                         PIL.Image.Image: _pil_to_b64,
                         pd.DataFrame: lambda x: x.to_dict(orient="records")}

    for class_ in _transform_mapper:
        if issubclass(subclass, class_):
            return _transform_mapper[class_]

    no_transform_fxn = lambda x: x

    return no_transform_fxn


def _transform_outputs(outputs, tags):
    "Transform outputs to fit in the desired components"

    return [_get_transform_function(o, tag)(o) for (o, tag) in zip(outputs, tags)]


def _transform_inputs(inputs, tags):
    "Transform inputs to fit in the desired components"

    transformed_inputs = []
    for inp, tag in zip(inputs, tags):
        if inp is None:
            transformed_inputs.append(inp)

        elif tag == "Image":
            transformed_inputs.append(_b64_to_pil(inp))

        else:
            transformed_inputs.append(inp)

    return transformed_inputs


def _get_error_notification_component(error_text):
    return dmc.Notification(
        title="Error",
        id="simple-notify",
        action="show",
        color="red",
        message=error_text.capitalize(),
        icon=DashIconify(icon="bxs:error"),
    )


def _clean_text(string):
    # Use regular expression to replace non-alphanumeric characters with underscores
    pattern_non_alpha_numeric = r"\W"  # \W matches any non-alphanumeric character
    replacement_non_alpha_numeric = "_"
    string = re.sub(pattern_non_alpha_numeric, replacement_non_alpha_numeric, string)

    # Use regular expression to remove consecutive underscores and boundary underscores
    pattern_consecutive_underscores = r"(^_+)|(_{2,})|(_+$)"
    replacement_consecutive_underscores = ""
    string = re.sub(
        pattern_consecutive_underscores, replacement_consecutive_underscores, string
    )

    return string.upper()


def _infer_variable_names(func):
    s = inspect.getsource(func)
    final_line = s.split("return")[-1].strip()
    line_without_comment = final_line.split("#")[0].strip().split(",")

    variable_names = [_clean_text(s) for s in line_without_comment]

    return variable_names


def _get_default_property(component_type):
    """
    Define a list of all important Dash components and their main
    properties to use as default.
    """

    try:
        default_component_attributes = {
            # Dash Core Components
            dcc.DatePickerSingle: "date",
            dcc.Dropdown: "value",
            dcc.Graph: "figure",
            dcc.Input: "value",
            dcc.Link: "href",
            dcc.Markdown: "children",
            dcc.RangeSlider: "value",
            dcc.RadioItems: "value",
            dcc.Slider: "value",
            dcc.Textarea: "value",
            dcc.Upload: "contents",
            # Dash Bootstrap Components
            dbc.Alert: "children",
            dbc.Badge: "children",
            dbc.Checkbox: "value",
            dbc.Col: "children",
            dbc.Collapse: "is_open",
            dbc.Container: "children",
            dbc.DropdownMenu: "children",
            dbc.Form: "children",
            dbc.Input: "value",
            dbc.ListGroup: "children",
            dbc.ListGroupItem: "children",
            dbc.Modal: "is_open",
            dbc.Row: "children",
            dbc.Table: "children",
            # Dash HTML Components
            html.H1: "children",
            html.H2: "children",
            html.H3: "children",
            html.H4: "children",
            html.H5: "children",
            html.H6: "children",
            html.I: "children",
            html.Iframe: "srcdoc",
            html.Img: "src",
            html.Table: "children",
            html.Tbody: "children",
            html.Td: "children",
            html.Textarea: "value",
            html.Tfoot: "children",
            html.Th: "children",
            html.Thead: "children",
            html.Title: "children",
            html.Video: "src",
            # Dash mantine components
            dmc.Checkbox: "checked",
            dmc.Chip: "checked",
            dmc.ChipGroup: "value",
            dmc.ColorPicker: "value",
            dmc.DatePicker: "value",
            dmc.DateRangePicker: "value",
            dmc.JsonInput: "value",
            dmc.MultiSelect: "value",
            dmc.NumberInput: "value",
            dmc.PasswordInput: "value",
            dmc.RadioGroup: "value",
            dmc.SegmentedControl: "value",
            dmc.Select: "value",
            dmc.Slider: "value",
            dmc.Switch: "checked",
            dmc.TextInput: "value",
            dmc.Textarea: "value",
            dmc.TimeInput: "value",
            dmc.Blockquote: "children",
            dmc.Code: "children",
            dmc.List: "children",
            dmc.Prism: "children",
            dmc.Text: "children",
            dmc.Title: "children",
        }

        default_property = default_component_attributes.get(component_type, "value")

    except Exception as e:
        warnings.warn(str(e))
        default_property = "value"

    return default_property


def _parse_docstring_as_markdown(func, title=None, get_short=False):
    """
    Generate markdown documentation from the docstring of a function.

    Args:
        func (function): The function to document.
        title (str, optional): Title of the text. If None, it is evaluated. Defaults to None.
        get_short (bool, optional): If True, returns only the short description. Defaults to False.

    Returns:
        str: Markdown documentation string.
    """
    logging.warning("Parsing function docstring is still an experimental feature. To reduce uncertainty, consider setting `about` to `False`.")

    parsed = docstring_parser.parse(func.__doc__)

    if get_short == True:
        return parsed.short_description

    # Start with the function description
    md_list = [
        f"#### {func.__name__.replace('_', ' ').title() if not title else title}",
        "",
        parsed.short_description or "",
        "",
        parsed.long_description or "",
        "",
    ]

    # Add parameters in table format
    if parsed.params:
        md_list.extend(
            [
                "##### Parameters",
                "| Name | Type | Mandatory | Default | Description |",
                "| ---- | ---- | --------- | ------- | ----------- |",
            ]
        )

        for param in parsed.params:
            param_line = (
                f"| {param.arg_name} "
                f"| {param.type_name or 'Not specified'} "
                f"| {'No' if param.is_optional else 'Yes'} "
                f"| {param.default or 'None'} "
                f"| {param.description or 'No description provided'} |"
            )
            md_list.append(param_line)

    md_list.extend(["# ", "# "])

    # Add return values in table format
    if parsed.returns:
        md_list.extend(
            ["", "##### Returns", "| Type | Description |", "| ---- | ----------- |"]
        )

        return_line = (
            f"| {parsed.returns.type_name or 'Not specified'} "
            f"| {parsed.returns.description or 'No description provided'} |"
        )
        md_list.append(return_line)

    # Join all and return markdown string
    return "\n".join(md_list)
